from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
)

import struct
from io import BytesIO
import os
from uuid import UUID, uuid4
from binascii import hexlify, unhexlify

class AVBObjectRef(object):
    __slots__ = ('root', 'index')
    def __init__(self, root, index):
        self.root = root
        self.index = index

    @property
    def value(self):
        if self.index <= 0:
            return None
        if self.root.debug_copy_refs:
            return self

        return self.root.read_object(self.index)

    @property
    def class_id(self):
        if not self.root.check_refs:
            return b'NULL'
        if self.valid:
            chunk = self.root.read_chunk(self.index)
            return chunk.class_id

    @property
    def valid(self):
        if self.index >= len(self.root.object_positions):
            return False
        return True

    def __repr__(self):
        s = "%s.%s"  % (self.__class__.__module__,
                                self.__class__.__name__)
        if self.root and self.index and self.valid:
            chunk = self.root.read_chunk(self.index)
            s += " %s idx: %d pos: %d" % (chunk.class_id, self.index, chunk.pos)
        return '<%s at 0x%x>' % (s, id(self))

def int_from_bytes(data, byte_order='big'):
    num = 0
    if byte_order == 'little':
        for i, byte in enumerate(data):
            num += byte << (i * 8)
        return num
    elif byte_order == 'big':
        length = len(data) - 1
        for i, byte in enumerate(data):
            num += byte << ((length-i) * 8)
        return num
    else:
        raise ValueError('endianess must be "little" or "big"')

def bytes_from_int(num, length, byte_order='big'):
    if byte_order == 'little':
        v = bytearray((num >> (i * 8) & 0xff for i in range(length)))
        return bytes(v)
    elif byte_order == 'big':
        v = bytearray((num >> (length - 1 - i) * 8) & 0xff for i in range(length))
        return bytes(v)
    else:
        raise ValueError('endianess must be "little" or "big"')

def peek_data(f, size=None):
    pos = f.tell()
    if size:
        data = f.read(size)
    else:
        data = f.read()
    f.seek(pos)
    return data

def unpack_u16le_from(buffer, offset):
    value  = buffer[offset]
    value += buffer[offset+1] << 8
    return value

def unpack_u32le_from(buffer, offset):
    value  = buffer[offset]
    value += buffer[offset+1] << 8
    value += buffer[offset+2] << 16
    value += buffer[offset+3] << 24
    return value

def generate_uid():
    v = uuid4().int & (1<<64)-1
    return v

AVBClaseID_dict = {}
AVBClassName_dict = {}
def register_class(classobj):
    AVBClaseID_dict[classobj.class_id] = classobj
    AVBClassName_dict[classobj.__name__] = classobj

    for pdef in classobj.propertydefs:
        classobj.propertydefs_dict[pdef.name] = pdef

    return classobj

def register_helper_class(classobj):
    AVBClassName_dict[classobj.__name__] = classobj
    for pdef in classobj.propertydefs:
        classobj.propertydefs_dict[pdef.name] = pdef
    return classobj



# Copied from pyaaf2

"""

Excerpt from SMPTE ST 330 (Focus on Basic UMID)::

    5 General Specification

    A unique material identifier (UMID) provides for the globally unique identification of any audiovisual material.
    This standard defines a dual approach through the specification of a basic UMID and an extended UMID.
    The basic UMID provides a globally unique identification for audiovisual material that comprises an integer
    number of one or more contiguous material units. The basic UMID has no embedded mechanism to
    distinguish between individual material units within a single instance of audiovisual material. The data in the
    basic UMID can be created through automatic generation.

    The extended UMID comprises the basic UMID followed immediately by a source pack that provides a
    signature for material units. The source pack comprises a fixed length metadata pack of 32 bytes that
    provides sufficient metadata by which source ?when, where and who (or what)? information can be identified
    regardless of current ownership or status. The extended UMID also provides a mechanism to distinguish
    between individual material units within a single instance of audiovisual material.

    The basic UMID is 32 bytes long and the extended UMID is 64 bytes long.
    Both UMID types use the key-length-value construct defined by SMPTE ST 336. The key is a 16-byte
    universal label truncated to 12 bytes.
    In the case of the basic UMID, the length field has a value of 13h and the value is formed by the combination
    of a material number and an instance number.
    In the case of the extended UMID, the length field has a value of 33h and the value is formed by the
    combination of the material and the instance numbers followed by the source pack.
    All components of the UMID have a defined byte order for consistent application in storage and streaming
    environments.

    The components of the basic UMID are:
    1. A 12-byte universal label,
    2. A 1-byte length value,
    3. A 3-byte instance number, and
    4. A 16-byte material number.

    The combination of the instance and material numbers can be treated as a dumb number.
    Note: The material number does not indicate the status of the material (such as copy number) or its representation
    (such as the compression kind). The material number can be identical in copies and in different representations of
    the material. The purpose of the instance number is to separately identify different representations or instances of
    audiovisual material. Thus, for example, a high-resolution picture and a thumbnail can both have the same
    material number because they both represent the same picture but, because they are different instances, they will
    have different instance numbers for the different representations. Guidance for the consistent application of new
    material numbers and instance numbers is given in SMPTE RP 205.


    UMID univeral label (SMPTELabel)

    Byte No.   Description            Value (hex)                 Meaning
    ----------------------------------------------------------------------------------------
    1          Object identifier      06h                         Universal label start
    2          Label size             0Ah                         12-byte Universal label
    3          Designation: ISO       2Bh                         ISO registered
    4          Designation: SMPTE     34h                         SMPTE registered
    5          Registry category      01h                         Dictionaries
    6          Specific category      01h                         Metadata dictionaries
    7          Structure              01h                         Dictionary standard (SMPTE ST 335)
    8          Version number         05h                         Version of the metadata dictionary (defined in SMPTE RP 210)
    9          Class                  01h                         Identifiers and locators
    10         Subclass               01h                         Globally unique identifiers
    11         Material type          XXh                         See Section 6.1.2.1
    12         Number creation method YYh                         See Section 6.1.2.2

    6.1.2.1 - Meterial type identification

    Byte 11 of the UL shall define the material type being identified using one of the values defined in Table 2.
    The use of material types '01h', '02h', '03h' and '04h' shall be deprecated for use in implementations using
    this revised standard. These values are preserved only for compatibility with systems implemented using
    SMPTE ST 330:2000#

    Table 2

    Byte value    Meaning                                                            Examples and notes
    --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    01h           picture material                                                   Deprecated
    02h           audio material                                                     Deprecated
    03h           data material                                                      Deprecated
    04h           other material                                                     Deprecated (originally not only picture, audio, or data material, but also a combination of material types)
    05h           single picture component                                           e.g. Y component
    06h           Two or more picture components in a single container               e.g. interleaved Y, Cb and Cr components
    08h           single audio component                                             e.g mono audio
    09h           two or more audio components in a single container                 e.g. AES3 audio pair
    0Bh           single auxiliary (or data) component                               e.g. sub-titles only
    0Ch           two or more auxiliary (or data) components in a single container   e.g. multiple sub-titles streams in different languages
    0Dh           mixed group of components in a single container                    e.g. video & stereo audio pair
    0Fh           material type is not identified

    6.1.2.2 Number creation method identification

    Byte 12 of the UL shall define the method by which the material and instance numbers are created. This byte
    is divided into top and bottom nibbles for the purpose of this definition.
    The top nibble shall occupy the 4 most significant bits (MSBs) of the byte and the value shall be used to
    define the method of material number creation. The values used by this nibble shall be limited to the range 0
    to 7h so that byte 12 conforms to the ASN.1 BER short form coding rules used by SMPTE ST 298.
    The methods of material number generation shall be as defined in table 3 and the specification of the each
    method shall be as defined in Annex A.
    Note: New material number generation methods can be added by amendment or revision of this document. Each
    addition will provide the proposed value (within the range of values currently identified as "Reserved but not
    defined") for inclusion in Table 3 together with the supporting definition to be added to Annex A.

    Table 3 - Identification of material number generation method::

        Value (hex)   Method
        -------------------------------------------------------------
        0             No defined method
        1             SMPTE method
        2             UUID/UL method
        3             Masked method
        4             IEEE 1394 network method
        5~7           Reserved but not defined


Notes from Pixar 10/30/17

Final note of discussion with Avid engineers how the top nibble in the 12th byte in the SMPTELabel should be set.
(In the past we always had it set to 00, i.e. "no defined method", we had some confusion about how to set it when using a uuid for the material)

Avid Engineer:

"The specification of number creation identification is very clear about using the 4 MSBs for the material
number so I am pretty sure that the numbers in the sub-titles of Annex A (e.g. 02h) should not be interpreted literally as values of byte 12.
20h is the correct value of the byte 12 for UL/UUID method/No defined method.

By the way, I don't see anything wrong with setting byte 12 to 00h (No defined method / No defined method)"

We at Pixar decided to set the byte to 20h, since it (even if already very minimal) completely eliminates the possibility to collide with any
MOB ID created by our old MOB ID generation algorithm.
"""

from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )

import uuid
import struct
from .utils import (int_from_bytes, bytes_from_int,
                    unpack_u16le_from, unpack_u32le_from)

MOBID_STRUCT = struct.Struct(str(''.join(( '<',
   '12B',  # UInt8Array12   SMPTELabel      0
   'B',    # UInt8          length         12
   'B',    # UInt8          instanceHigh   13
   'B',    # UInt8          instanceMid    14
   'B',    # UInt8          instanceLow    15
   'I',    # UInt32         Data1          16
   'H',    # UInt16         Data2          20
   'H',    # UInt16         Data3          22
   '8B',   # UInt8Array8    Data4          24
 ))))
def UniqueMobID():
    m = MobID()              # Description              Meaning
    m.SMPTELabel = [0x06,    # Object identifier        Universal label start
                    0x0a,    # Label size               12-byte Universal label
                    0x2b,    # Designation: ISO         ISO registered
                    0x34,    # Designation: SMPTE       SMPTE registered
                    0x01,    # Registry category        Dictionaries
                    0x01,    # Specific category        Metadata dictionaries
                    0x01,    # Structure                Dictionary standard (SMPTE ST 335)
                    0x05,    # Version number           Version of the metadata dictionary (defined in SMPTE RP 210)
                    0x01,    # Class                    Identifiers and locators
                    0x01,    # Subclass                 Globally unique identifiers
                    0x0f,    # Material type            See Section 6.1.2.1
                    0x20     # Number creation method   See Section 6.1.2.2     # Using UUID/UL method, Note 10/30/17 - matching pixar recommendation
                   ]
    m.length = 0x13          # Length, 13h = Basic UMID, 33h = Extended UMID
    m.instanceHigh = 0x00
    m.instanceMid = 0x00
    m.instanceLow = 0x00
    m.material = uuid.uuid4() # 16 byte material slot, filled with uuid according to RFC4122
    return m

class MobID(object):
    __slots__ = ('bytes_le')
    def __init__(self, mobid=None, bytes_le=None, int=None):

        if bytes_le:
            self.bytes_le = bytearray(bytes_le)
        else:
            self.bytes_le = bytearray(32)

            if mobid is not None:
                self.urn = mobid

            if int is not None:
                self.int = int

    @staticmethod
    def new():
        """
        Static method for generating unique MobIDs. Uses uuid.uuid4() for generation.
        """
        return UniqueMobID()

    @property
    def material(self):
        """
        MobID material representation as a UUID
        """
        return uuid.UUID(bytes_le=bytes(self.bytes_le[16:]))

    @material.setter
    def material(self, value):
        self.bytes_le[16:] = value.bytes_le

    @property
    def SMPTELabel(self):
        return self.bytes_le[0:12]

    @SMPTELabel.setter
    def SMPTELabel(self, value):
        struct.pack_into(str('12B'), self.bytes_le, 0, *value)

    @property
    def length(self):
        return self.bytes_le[12]

    @length.setter
    def length(self, value):
        self.bytes_le[12] = value

    @property
    def instanceHigh(self):
        return self.bytes_le[13]

    @instanceHigh.setter
    def instanceHigh(self, value):
        self.bytes_le[13] = value

    @property
    def instanceMid(self):
        return self.bytes_le[14]

    @instanceMid.setter
    def instanceMid(self, value):
        self.bytes_le[14] = value

    @property
    def instanceLow(self):
        return self.bytes_le[15]

    @instanceLow.setter
    def instanceLow(self, value):
        self.bytes_le[15] = value

    @property
    def Data1(self):
        return unpack_u32le_from(self.bytes_le, 16)

    @Data1.setter
    def Data1(self, value):
        struct.pack_into(str('<I'), self.bytes_le, 16, value)

    @property
    def Data2(self):
        return unpack_u16le_from(self.bytes_le, 20)

    @Data2.setter
    def Data2(self, value):
        struct.pack_into(str('<H'), self.bytes_le, 20, value)

    @property
    def Data3(self):
        return unpack_u16le_from(self.bytes_le, 22)

    @Data3.setter
    def Data3(self, value):
        struct.pack_into(str('<H'), self.bytes_le, 22, value)

    @property
    def Data4(self):
        return self.bytes_le[24:32]

    @Data4.setter
    def Data4(self, value):
        struct.pack_into(str('8B'), self.bytes_le, 24, *value)

    def from_dict(self, d):
        """
        Set MobID from a dict
        """
        self.length = d.get("length", 0)
        self.instanceHigh = d.get("instanceHigh", 0)
        self.instanceMid = d.get("instanceMid", 0)
        self.instanceLow = d.get("instanceLow", 0)

        material = d.get("material", {'Data1':0, 'Data2':0, 'Data3':0, 'Data4': [0 for i in range(8)]})

        self.Data1 = material.get('Data1', 0)
        self.Data2 = material.get('Data2', 0)
        self.Data3 = material.get('Data3', 0)

        self.Data4 = material.get("Data4", [0 for i in range(8)])
        self.SMPTELabel = d.get("SMPTELabel", [0 for i in range(12)])

    def to_dict(self):
        """
        MobID representation as dict
        """

        material = {'Data1': self.Data1,
                    'Data2': self.Data2,
                    'Data3': self.Data3,
                    'Data4': list(self.Data4)
                    }

        return {'material':material,
                'length': self.length,
                'instanceHigh': self.instanceHigh,
                'instanceMid': self.instanceMid,
                'instanceLow': self.instanceLow,
                'SMPTELabel': list(self.SMPTELabel)
                }
    @property
    def int(self):
        """
        MobID representation as a int
        """
        return int_from_bytes(self.bytes_le, byte_order='big')

    @int.setter
    def int(self, value):
        # NOTE: interpreted as big endian
        self.bytes_le = bytearray(bytes_from_int(value, 32, byte_order='big'))

    def __int__(self):
        return self.int

    def __eq__(self, other):
        if isinstance(other, MobID):
            return self.bytes_le == other.bytes_le
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, MobID):
            return self.bytes_le != other.bytes_le
        return NotImplemented

    def __lt__(self, other):
        if isinstance(other, MobID):
            return self.int < other.int
        return NotImplemented

    def __le__(self, other):
        if isinstance(other, MobID):
            return self.int <= other.int
        return NotImplemented

    def __gt__(self, other):
        if isinstance(other, MobID):
            return self.int > other.int
        return NotImplemented

    def __ge__(self, other):
        if isinstance(other, MobID):
            return self.int >= other.int
        return NotImplemented

    def __hash__(self):
        return hash(bytes(self.bytes_le))

    @property
    def urn(self):
        """
        MobID Uniform Resource Name representation.
        https://en.wikipedia.org/wiki/Uniform_Resource_Name
        """

        SMPTELabel = self.SMPTELabel
        Data4 = self.Data4

        # handle case UMIDs where the material number is half swapped
        if (SMPTELabel[11] == 0x00 and
            Data4[0] == 0x06 and Data4[1] == 0x0E and
            Data4[2] == 0x2B and Data4[3] == 0x34 and
            Data4[4] == 0x7F and Data4[5] == 0x7F):

            # print("case 1")
            f = "urn:smpte:umid:%02x%02x%02x%02x.%02x%02x%02x%02x.%02x%02x%02x%02x." + \
             "%02x"  + \
             "%02x%02x%02x." + \
             "%02x%02x%02x%02x.%02x%02x%02x%02x.%08x.%04x%04x"

            return f % (
                 SMPTELabel[0], SMPTELabel[1], SMPTELabel[2],  SMPTELabel[3],
                 SMPTELabel[4], SMPTELabel[5], SMPTELabel[6],  SMPTELabel[7],
                 SMPTELabel[8], SMPTELabel[9], SMPTELabel[10], SMPTELabel[11],
                 self.length,
                 self.instanceHigh, self.instanceMid, self.instanceLow,
                 Data4[0], Data4[1], Data4[2], Data4[3],
                 Data4[4], Data4[5], Data4[6], Data4[7],
                 self.Data1, self.Data2, self.Data3)
        else:
            # print("case 2")
            f = "urn:smpte:umid:%02x%02x%02x%02x.%02x%02x%02x%02x.%02x%02x%02x%02x." + \
             "%02x"  + \
             "%02x%02x%02x." + \
             "%08x.%04x%04x.%02x%02x%02x%02x.%02x%02x%02x%02x"

            return f % (
                 SMPTELabel[0], SMPTELabel[1], SMPTELabel[2],  SMPTELabel[3],
                 SMPTELabel[4], SMPTELabel[5], SMPTELabel[6],  SMPTELabel[7],
                 SMPTELabel[8], SMPTELabel[9], SMPTELabel[10], SMPTELabel[11],
                 self.length,
                 self.instanceHigh, self.instanceMid, self.instanceLow,
                 self.Data1, self.Data2, self.Data3,
                 Data4[0], Data4[1], Data4[2], Data4[3],
                 Data4[4], Data4[5], Data4[6], Data4[7])

    @urn.setter
    def urn(self, value):
        s = str(value).lower()
        for item in ("urn:smpte:umid:", ".", '-', '0x'):
            s = s.replace(item, '')
        assert len(s) == 64

        SMPTELabel = [0 for i in range(12)]
        start = 0
        for i in range(12):
            end = start + 2
            v = s[start:end]
            SMPTELabel[i] = int(v, 16)
            start = end
        self.SMPTELabel = SMPTELabel
        self.length = int(s[24:26], 16)
        self.instanceHigh = int(s[26:28], 16)
        self.instanceMid = int(s[28:30], 16)
        self.instanceLow = int(s[30:32], 16)

        start = 32
        data = [0 for i in range(6)]
        for i in range(6):
            end = start + 2
            v = s[start:end]
            data[i] = int(v, 16)
            start = end
        # print(s[32:start])
        if (SMPTELabel[11] == 0x00 and
                   data[0] == 0x06 and data[1] == 0x0E and
                   data[2] == 0x2B and data[3] == 0x34 and
                   data[4] == 0x7F and data[5] == 0x7F):

            start = 32
            data4 = [0 for i in range(8)]
            for i in range(8):
                end = start + 2
                v = s[start:end]
                data4[i] = int(v, 16)
                start = end

            self.Data4 = data4
            self.Data1 = int(s[48:56], 16)
            self.Data2 = int(s[56:60], 16)
            self.Data3 = int(s[60:64], 16)

        else:
            self.Data1 = int(s[32:40], 16)
            self.Data2 = int(s[40:44], 16)
            self.Data3 = int(s[44:48], 16)

            start = 48
            data4 = [0 for i in range(8)]
            for i in range(8):
                end = start + 2
                v = s[start:end]
                data4[i] = int(v, 16)
                start = end
            self.Data4 = data4

    def __repr__(self):
        return str(self.urn)

if __name__ == "__main__":

    t = "urn:smpte:umid:060a2b34.01010101.01010f00.13000000.060e2b34.7f7f2a80.4fa5c20f.4e301e50"
    m = MobID(t)
    print(t)
    print(m)
    assert str(m) == t



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
import datetime

from . import core
from .core import AVBPropertyDef, AVBRefList
from .components import Component
from . import utils
from . import mobid
from . utils import peek_data

TRACK_LABEL_FLAG            = 1 << 0
TRACK_ATTRIBUTES_FLAG       = 1 << 1
TRACK_COMPONENT_FLAG        = 1 << 2
TRACK_FILLER_PROXY_FLAG     = 1 << 3
TRACK_BOB_DATA_FLAG         = 1 << 4
TRACK_CONTROL_CODE_FLAG     = 1 << 5
TRACK_CONTROL_SUB_CODE_FLAG = 1 << 6
TRACK_START_POS_FLAG        = 1 << 7
TRACK_READ_ONLY_FLAG        = 1 << 8
TRACK_SESSION_ATTR_FLAG     = 1 << 9

TRACK_UNKNOWN_FLAGS         = 0xFC00

@utils.register_helper_class
class Track(core.AVBObject):
    propertydefs_dict = {}
    propertydefs = [
        AVBPropertyDef('index',            'OMFI:TRAK:LabelNumber',    'int16'),
        AVBPropertyDef('attributes',       'OMFI:TRAK:Attributes',     'reference'),
        AVBPropertyDef('session_attr',     'OMFI:TRAK:SessionAttrs',   'reference'),
        AVBPropertyDef('component',        'OMFI:TRAK:TrackComponent', 'reference'),
        AVBPropertyDef('filler_proxy',     'OMFI:TRAK:FillerProxy',    'reference'),
        AVBPropertyDef('bob_data',         '__OMFI:TRAK:Bob',          'reference'),
        AVBPropertyDef('control_code',     'OMFI:TRAK:ControlCode',    'int16'),
        AVBPropertyDef('control_sub_code', 'OMFI:TRAK:ControlSubCode', 'int16'),
        AVBPropertyDef('start_pos',        'OMFI:TRAK:StartPos',       'int32'),
        AVBPropertyDef('read_only',        '__OMFI:TRAK:ReadOnly',     'bool'),
        AVBPropertyDef('lock_number',      'OMFI:TRAK:LockNubmer',     'int16'),
    ]
    __slots__ = ()

    @property
    def media_kind(self):
        if hasattr(self, 'component'):
            return self.component.media_kind

    @property
    def flags(self):
        flags = 0
        if 'index' in self.property_data:
            flags |= TRACK_LABEL_FLAG

        if 'attributes' in self.property_data:
            flags |= TRACK_ATTRIBUTES_FLAG

        if 'session_attr' in self.property_data:
            flags |= TRACK_SESSION_ATTR_FLAG

        if 'component' in self.property_data:
            flags |= TRACK_COMPONENT_FLAG

        if 'filler_proxy' in self.property_data:
            flags |= TRACK_FILLER_PROXY_FLAG

        if 'bob_data' in self.property_data:
            flags |= TRACK_BOB_DATA_FLAG

        if 'control_code' in self.property_data:
            flags |= TRACK_CONTROL_CODE_FLAG

        if 'control_sub_code' in self.property_data:
            flags |= TRACK_CONTROL_SUB_CODE_FLAG

        if 'start_pos' in self.property_data:
            flags |= TRACK_START_POS_FLAG

        if 'read_only' in self.property_data:
            flags |= TRACK_READ_ONLY_FLAG

        return flags


@utils.register_class
class TrackGroup(Component):
    class_id = b'TRKG'
    propertydefs_dict = {}
    propertydefs = Component.propertydefs + [
        AVBPropertyDef('mc_mode',     'OMFI:TRKG:MC:Mode',     'int8',    0),
        AVBPropertyDef('length',      'OMFI:TRKG:GroupLength',  'int32',  0),
        AVBPropertyDef('num_scalars', 'OMFI:TRKG:NumScalars',  'int32',   0),
        AVBPropertyDef('tracks',      'OMFI:TRKG:Tracks',      'list',     )
    ]
    __slots__ = ()

    def __init__(self, edit_rate=25, media_kind=None):
        super(TrackGroup, self).__init__(edit_rate=edit_rate, media_kind=media_kind)
        self.tracks = []

    def read(self, f):
        super(TrackGroup, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x08)

        self.mc_mode = ctx.read_u8(f)
        self.length = ctx.read_s32(f)
        self.num_scalars = ctx.read_s32(f)

        track_count = ctx.read_s32(f)
        self.tracks = []
        # print("tracks:", track_count)
        # really annoying, tracks can have variable lengths!!

        for i in range(track_count):
            # print(peek_data(f).encode("hex"))
            track = Track.__new__(Track, root=self.root)
            flags = ctx.read_u16(f)

            if flags & TRACK_LABEL_FLAG:
                track.index = ctx.read_s16(f)

            if flags & TRACK_ATTRIBUTES_FLAG:
                track.attributes = ctx.read_object_ref(self.root, f)

            if flags & TRACK_SESSION_ATTR_FLAG:
                track.session_attr = ctx.read_object_ref(self.root, f)

            if flags & TRACK_COMPONENT_FLAG:
                track.component = ctx.read_object_ref(self.root, f)

            if flags & TRACK_FILLER_PROXY_FLAG:
                track.filler_proxy = ctx.read_object_ref(self.root, f)

            if flags & TRACK_BOB_DATA_FLAG:
                track.bob_data = ctx.read_object_ref(self.root, f)

            if flags & TRACK_CONTROL_CODE_FLAG:
                track.control_code = ctx.read_s16(f)

            if flags & TRACK_CONTROL_SUB_CODE_FLAG:
                track.control_sub_code = ctx.read_s16(f)

            if flags & TRACK_START_POS_FLAG:
                track.start_pos = ctx.read_s32(f)

            if flags & TRACK_READ_ONLY_FLAG:
                track.read_only = ctx.read_bool(f)

            if flags & TRACK_UNKNOWN_FLAGS:
                raise ValueError("Unknown Track Flag: %d" % flags)

            assert track.flags == flags

            self.tracks.append(track)

        for tag in ctx.iter_ext(f):
            if tag == 0x01:
                for i in range(track_count):
                    ctx.read_assert_tag(f, 69)
                    self.tracks[i].lock_number = ctx.read_s16(f)
            else:
                raise ValueError("%s: unknown ext tag 0x%02X %d" % (str(self.class_id), tag,tag))

    def write(self, f):
        super(TrackGroup, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x08)

        ctx.write_u8(f, self.mc_mode)
        ctx.write_s32(f, self.length)
        ctx.write_s32(f, self.num_scalars)

        ctx.write_s32(f, len(self.tracks))

        for track in self.tracks:
            flags = track.flags
            if flags & TRACK_UNKNOWN_FLAGS:
                raise ValueError("Unknown Track Flag: %d" % flags)

            ctx.write_u16(f, flags)
            if flags & TRACK_LABEL_FLAG:
                ctx.write_s16(f, track.index)

            if flags & TRACK_ATTRIBUTES_FLAG:
                ctx.write_object_ref(self.root, f, track.attributes)

            if flags & TRACK_SESSION_ATTR_FLAG:
                ctx.write_object_ref(self.root, f, track.session_attr)

            if flags & TRACK_COMPONENT_FLAG:
                ctx.write_object_ref(self.root, f, track.component)

            if flags & TRACK_FILLER_PROXY_FLAG:
                ctx.write_object_ref(self.root, f, track.filler_proxy)

            if flags & TRACK_BOB_DATA_FLAG:
                ctx.write_object_ref(self.root, f, track.bob_data)

            if flags & TRACK_CONTROL_CODE_FLAG:
                ctx.write_s16(f, track.control_code)

            if flags & TRACK_CONTROL_SUB_CODE_FLAG:
                ctx.write_s16(f, track.control_sub_code)

            if flags & TRACK_START_POS_FLAG:
                ctx.write_s32(f, track.start_pos)

            if flags & TRACK_READ_ONLY_FLAG:
                ctx.write_bool(f, track.read_only)

        if self.tracks:
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x01)

            for i in range(len(self.tracks)):
                ctx.write_u8(f, 69)
                if hasattr(self.tracks[i], 'lock_number'):
                    ctx.write_s16(f, self.tracks[i].lock_number)
                else:
                    ctx.write_s16(f, 0)


@utils.register_class
class TrackEffect(TrackGroup):
    class_id = b'TKFX'
    propertydefs_dict = {}
    propertydefs = TrackGroup.propertydefs + [
        AVBPropertyDef('left_length',         'OMFI:TKFX:MC:LeftLength',            'int32'),
        AVBPropertyDef('right_length',        'OMFI:TKFX:MC:RightLength',           'int32'),
        AVBPropertyDef('info_version',        'OMFI:TNFX:MC:GlobalInfoVersion',     'int16'),
        AVBPropertyDef('info_current',        'OMFI:TNFX:MC:GlobalInfo.kfCurrent',  'int32'),
        AVBPropertyDef('info_smooth',         'OMFI:TNFX:MC:GlobalInfo.kfSmooth',   'int32'),
        AVBPropertyDef('info_color_item',     'OMFI:TNFX:MC:GlobalInfo.colorItem',  'int16'),
        AVBPropertyDef('info_quality',        'OMFI:TNFX:MC:GlobalInfo.quality',    'int16'),
        AVBPropertyDef('info_is_reversed',    'OMFI:TNFX:MC:GlobalInfo.isReversed', 'int8'),
        AVBPropertyDef('info_aspect_on',      'OMFI:TNFX:MC:GlobalInfo.aspectOn',   'bool'),
        AVBPropertyDef('keyframes',           'OMFI:TKFX:MC:KeyFrameList',     'reference'),
        AVBPropertyDef('info_force_software', 'OMFI:TNFX:MC:ForceSoftware',         'bool'),
        AVBPropertyDef('info_never_hardware', 'OMFI:TKFX:MC:NeverHardware',         'bool'),
        AVBPropertyDef('trackman',            'OMFI:TKFX:MC:TrackMan',         'reference'),
    ]
    __slots__ = ()

    def read(self, f):
        super(TrackEffect, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x06)

        self.left_length = ctx.read_s32(f)
        self.right_length = ctx.read_s32(f)

        self.info_version = ctx.read_s16(f)
        self.info_current = ctx.read_s32(f)
        self.info_smooth = ctx.read_s32(f)
        self.info_color_item = ctx.read_s16(f)
        self.info_quality = ctx.read_s16(f)
        self.info_is_reversed = ctx.read_s8(f)
        self.info_aspect_on = ctx.read_bool(f)

        self.keyframes = ctx.read_object_ref(self.root, f)
        self.info_force_software = ctx.read_bool(f)
        self.info_never_hardware = ctx.read_bool(f)

        for tag in ctx.iter_ext(f):
            if tag == 0x02:
                ctx.read_assert_tag(f, 72)
                self.trackman = ctx.read_object_ref(self.root, f)
            else:
                raise ValueError("%s: unknown ext tag 0x%02X %d" % (str(self.class_id), tag,tag))

        if self.class_id[:] == b'TKFX':
            ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(TrackEffect, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x06)

        ctx.write_s32(f, self.left_length)
        ctx.write_s32(f, self.right_length)

        ctx.write_s16(f, self.info_version)
        ctx.write_s32(f, self.info_current)
        ctx.write_s32(f, self.info_smooth)
        ctx.write_s16(f, self.info_color_item)
        ctx.write_s16(f, self.info_quality)
        ctx.write_s8(f, self.info_is_reversed)
        ctx.write_bool(f, self.info_aspect_on)

        ctx.write_object_ref(self.root, f, self.keyframes)
        ctx.write_bool(f, self.info_force_software)
        ctx.write_bool(f, self.info_never_hardware)

        if hasattr(self, 'trackman'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x02)
            ctx.write_u8(f, 72)
            ctx.write_object_ref(self.root, f, self.trackman)

        if self.class_id[:] == b'TKFX':
            ctx.write_u8(f, 0x03)

@utils.register_class
class PanVolumeEffect(TrackEffect):
    class_id = b'PVOL'
    propertydefs_dict = {}
    propertydefs = TrackEffect.propertydefs + [
        AVBPropertyDef('level',                  'OMFI:PVOL:MC:Level',               'int32'),
        AVBPropertyDef('pan',                    'OMFI:PVOL:MC:Pan',                 'int32'),
        AVBPropertyDef('suppress_validation',    'OMFI:PVOL:MC:SuppressValidation',  'bool'),
        AVBPropertyDef('level_set',              'OMFI:PVOL:MC:LevelSet',            'bool'),
        AVBPropertyDef('pan_set',                'OMFI:PVOL:MC:PanSet',              'bool'),
        AVBPropertyDef('supports_seperate_gain', 'OMFI:PVOL:MC:DoesSuprtSeprtClipG', 'int32'),
        AVBPropertyDef('is_trim_gain_effect',    'OMFI:PVOL:MC:IsTrimGainEffect',    'int32'),
    ]
    __slots__ = ()

    def read(self, f):
        super(PanVolumeEffect, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x05)

        self.level = ctx.read_s32(f)
        self.pan = ctx.read_s32(f)

        self.suppress_validation = ctx.read_bool(f)
        self.level_set = ctx.read_bool(f)
        self.pan_set = ctx.read_bool(f)

        for tag in ctx.iter_ext(f):
            if tag == 0x01:
                ctx.read_assert_tag(f, 71)
                self.supports_seperate_gain = ctx.read_s32(f)
            elif tag == 0x02:
                ctx.read_assert_tag(f, 71)
                self.is_trim_gain_effect = ctx.read_s32(f)
            else:
                raise ValueError("%s: unknown ext tag 0x%02X %d" % (str(self.class_id), tag,tag))

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(PanVolumeEffect, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x05)

        ctx.write_s32(f, self.level)
        ctx.write_s32(f, self.pan)

        ctx.write_bool(f, self.suppress_validation)
        ctx.write_bool(f, self.level_set)
        ctx.write_bool(f, self.pan_set)

        if hasattr(self, 'supports_seperate_gain'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.supports_seperate_gain)
        if hasattr(self, 'is_trim_gain_effect'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x02)
            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.is_trim_gain_effect)

        ctx.write_u8(f, 0x03)

@utils.register_helper_class
class ASPIPlugin(core.AVBObject):
    propertydefs_dict = {}
    propertydefs = [
        AVBPropertyDef('name',             'OMFI:ASPI:plugInName',             'string'),
        AVBPropertyDef('manufacturer_id',  'OMFI:ASPI:plugInfManufacturerID',  'uint32'),
        AVBPropertyDef('product_id',       'OMFI:ASPI:plugInfProductID',       'uint32'),
        AVBPropertyDef('plugin_id',        'OMFI:ASPI:plugInfPlugInID',        'uint32'),
        AVBPropertyDef('chunks',           'OMFI:ASPI:plugInChunks',           'list'),
    ]
    __slots__ = ()

    def __init__(self, root):
        super(ASPIPlugin, self).__init__(root)
        self.chunks = []

@utils.register_helper_class
class ASPIPluginChunk(core.AVBObject):
    propertydefs_dict = {}
    propertydefs = [
        AVBPropertyDef('version',         'OMFI:ASPI:chunkfVersion',         'int32'),
        AVBPropertyDef('manufacturer_id', 'OMFI:ASPI:plugInfManufacturerID', 'uint32'),
        AVBPropertyDef('product_id',      'OMFI:ASPI:plugInfProductID',      'uint32'),
        AVBPropertyDef('plugin_id',       'OMFI:ASPI:plugInfPlugInID',       'uint32'),
        AVBPropertyDef('chunk_id',        'OMFI:ASPI:chunkfChunkID',         'uint32'),
        AVBPropertyDef('name',            'OMFI:ASPI:chunkfChunkName',       'string'),
        AVBPropertyDef('data',            'OMFI:ASPI:chunkfData',            'bytes'),
    ]
    __slots__ = ()

@utils.register_class
class AudioSuitePluginEffect(TrackEffect):
    class_id = b'ASPI'
    propertydefs_dict = {}
    propertydefs = TrackEffect.propertydefs + [
        AVBPropertyDef('plugins',          'OMFI:ASPI:plugIns',                         'list'),
        AVBPropertyDef('mob_id',           'MobID',                                     'MobID'),
        AVBPropertyDef('mark_in',          'OMFI:ASPI:markInForSourceMasterClip',       'uint64'),
        AVBPropertyDef('mark_out',         'OMFI:ASPI:markOutForSourceMasterClip',      'uint64'),
        AVBPropertyDef('tracks_to_affect', 'OMFI:ASPI:tracksToAffect',                  'uint32'),
        AVBPropertyDef('rendering_mode',   'OMFI:ASPI:renderingMode',                   'int32'),
        AVBPropertyDef('padding_secs',     'OMFI:ASPI:paddingSecs',                     'int32'),
        AVBPropertyDef('preset_path',      'OMFI:ASPI:presetPath',                      'bytes'),
    ]
    __slots__ = ()

    def read(self, f):
        super(AudioSuitePluginEffect, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        self.plugins = []

        number_of_plugins = ctx.read_s32(f)

        #TODO: find sample with multiple plugins
        assert number_of_plugins == 1

        plugin = ASPIPlugin.__new__(ASPIPlugin, root=self.root)
        plugin.name = ctx.read_string(f)
        plugin.manufacturer_id = ctx.read_u32(f)
        plugin.product_id = ctx.read_u32(f)
        plugin.plugin_id = ctx.read_u32(f)
        # print(peek_data(f).encode("hex"))
        num_of_chunks = ctx.read_s32(f)

        #TODO: find sample with multiple chunks
        # print('chunks', num_of_chunks)
        assert num_of_chunks >= 0
        plugin.chunks = []
        for i in range(num_of_chunks):

            chunk_size = ctx.read_s32(f)
            assert chunk_size >= 0

            chunk = ASPIPluginChunk.__new__(ASPIPluginChunk, root=self.root)
            chunk.version = ctx.read_s32(f)
            chunk.manufacturer_id = ctx.read_u32(f)
            chunk.product_id = ctx.read_u32(f)
            chunk.plugin_id = ctx.read_u32(f)

            chunk.chunk_id = ctx.read_u32(f)
            chunk.name = ctx.read_string(f)

            chunk.data = bytearray(f.read(chunk_size))

            plugin.chunks.append(chunk)

        self.plugins.append(plugin)

        # print(peek_data(f).encode("hex"))


        for tag in ctx.iter_ext(f):
            if tag == 0x01:
                # not sure what is used for. skiping
                ctx.read_assert_tag(f, 71)
                mob_hi = ctx.read_s32(f)
                ctx.read_assert_tag(f, 71)
                mob_lo = ctx.read_s32(f)
            elif tag == 0x02:
                ctx.read_assert_tag(f, 77)
                self.mark_in = ctx.read_s64(f)
            elif tag == 0x03:
                ctx.read_assert_tag(f, 77)
                self.mark_out = ctx.read_s64(f)
            elif tag == 0x04:
                ctx.read_assert_tag(f, 72)
                self.tracks_to_affect = ctx.read_s32(f)
            elif tag == 0x05:
                ctx.read_assert_tag(f, 71)
                self.rendering_mode = ctx.read_s32(f)
            elif tag == 0x06:
                ctx.read_assert_tag(f, 71)
                self.padding_secs = ctx.read_s32(f)
            elif tag == 0x08:
                self.mob_id = ctx.read_mob_id(f)

            elif tag == 0x09:
                ctx.read_assert_tag(f, 72)
                preset_path_length = ctx.read_u32(f)
                if preset_path_length > 0:
                    ctx.read_assert_tag(f, 65)
                    length = ctx.read_u32(f)
                    assert preset_path_length == length
                    self.preset_path = bytearray(f.read(length))
                else:
                    self.preset_path = bytearray()
            else:
                raise ValueError("%s: unknown ext tag 0x%02X %d" % (str(self.class_id), tag,tag))

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(AudioSuitePluginEffect, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)

        ctx.write_s32(f, len(self.plugins))
        for plugin in self.plugins:
            ctx.write_string(f, plugin.name)
            ctx.write_u32(f, plugin.manufacturer_id)
            ctx.write_u32(f, plugin.product_id)
            ctx.write_u32(f, plugin.plugin_id)

            ctx.write_s32(f, len(plugin.chunks))

            for chunk in plugin.chunks:

                ctx.write_s32(f, len(chunk.data))

                ctx.write_s32(f, chunk.version)
                ctx.write_u32(f, chunk.manufacturer_id)
                ctx.write_u32(f, chunk.product_id)
                ctx.write_u32(f, chunk.plugin_id)

                ctx.write_u32(f, chunk.chunk_id)
                ctx.write_string(f, chunk.name)

                f.write(chunk.data)

        if hasattr(self, 'mob_id'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x01)

            mob_lo = self.mob_id.material.time_low
            mob_hi = self.mob_id.material.time_mid + (self.mob_id.material.time_hi_version << 16)
            ctx.write_u8(f, 71)
            ctx.write_s32(f, mob_lo)
            ctx.write_u8(f, 71)
            ctx.write_s32(f, mob_hi)

        # NOTE: out of order to match seen files
        if hasattr(self, 'mob_id'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x08)
            ctx.write_mob_id(f, self.mob_id)

        if hasattr(self, 'mark_in'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x02)
            ctx.write_u8(f, 77)
            ctx.write_s64(f, self.mark_in)
        if hasattr(self, 'mark_out'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x03)
            ctx.write_u8(f, 77)
            ctx.write_s64(f, self.mark_out)
        if hasattr(self, 'tracks_to_affect'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x04)
            ctx.write_u8(f, 72)
            ctx.write_s32(f, self.tracks_to_affect)
        if hasattr(self, 'rendering_mode'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x05)
            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.rendering_mode)
        if hasattr(self, 'padding_secs'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x06)
            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.padding_secs)
        if hasattr(self, 'preset_path'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x09)
            ctx.write_u8(f, 72)
            # yes its twice for some reason
            ctx.write_u32(f, len(self.preset_path))
            if len(self.preset_path) > 0:
                ctx.write_u8(f, 65)
                ctx.write_u32(f, len(self.preset_path))
                f.write(self.preset_path)

        ctx.write_u8(f, 0x03)

@utils.register_helper_class
class EqualizerBand(core.AVBObject):
    propertydefs_dict = {}
    propertydefs = [
        AVBPropertyDef('type',   'OMFI:EQBD:AV:BandType',   'int32'),
        AVBPropertyDef('freq',   'OMFI:EQBD:AV:BandFreq',   'int32'),
        AVBPropertyDef('gain',   'OMFI:EQBD:AV:BandGain',   'int32'),
        AVBPropertyDef('q',      'OMFI:EQBD:AV:BandQ',      'int32'),
        AVBPropertyDef('enable', 'OMFI:EQBD:AV:BandEnable', 'bool'),
    ]
    __slots__ = ()

@utils.register_class
class EqualizerMultiBand(TrackEffect):
    class_id = b'EQMB'
    propertydefs_dict = {}
    propertydefs = TrackEffect.propertydefs + [
        AVBPropertyDef('bands',         'OMFI:EQBD:AV:Bands',        'list'),
        AVBPropertyDef('effect_enable', 'OMFI:EQMB:AV:EffectEnable', 'bool'),
        AVBPropertyDef('filter_name',   'OMFI:EQMB:AV:FilterName',   'string'),
    ]
    __slots__ = ()

    def read(self, f):
        super(EqualizerMultiBand, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x05)

        num_bands = ctx.read_s32(f)
        assert num_bands >= 0

        self.bands = []
        for i in range(num_bands):
            band = EqualizerBand.__new__(EqualizerBand, root=self.root)
            band.type = ctx.read_s32(f)
            band.freq = ctx.read_s32(f)
            band.gain = ctx.read_s32(f)
            band.q = ctx.read_s32(f)
            band.enable = ctx.read_bool(f)
            self.bands.append(band)

        self.effect_enable = ctx.read_bool(f)
        self.filter_name = ctx.read_string(f)

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(EqualizerMultiBand, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x05)

        ctx.write_s32(f, len(self.bands))
        for band in self.bands:
            ctx.write_s32(f, band.type)
            ctx.write_s32(f, band.freq)
            ctx.write_s32(f, band.gain)
            ctx.write_s32(f, band.q)
            ctx.write_bool(f, band.enable)

        ctx.write_bool(f, self.effect_enable)
        ctx.write_string(f, self.filter_name)

        ctx.write_u8(f, 0x03)

class TimeWarp(TrackGroup):
    class_id = b'WARP'
    propertydefs_dict = {}
    propertydefs = TrackGroup.propertydefs + [
        AVBPropertyDef('phase_offset', 'OMFI:WARP:PhaseOffset', 'int32'),
    ]
    __slots__ = ()

    def read(self, f):
        super(TimeWarp, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x02)
        self.phase_offset = ctx.read_s32(f)

    def write(self, f):
        super(TimeWarp, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x02)
        ctx.write_s32(f, self.phase_offset)

@utils.register_class
class CaptureMask(TimeWarp):
    class_id = b'MASK'
    propertydefs_dict = {}
    propertydefs = TimeWarp.propertydefs + [
        AVBPropertyDef('is_double', 'OMFI:MASK:IsDouble', 'bool'),
        AVBPropertyDef('mask_bits', 'OMFI:MASK:MaskBits', 'int32'),
    ]
    __slots__ = ()

    def read(self, f):
        super(CaptureMask, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        self.is_double = ctx.read_bool(f)
        self.mask_bits = ctx.read_u32(f)

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(CaptureMask, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)

        ctx.write_bool(f, self.is_double)
        ctx.write_u32(f, self.mask_bits)

        ctx.write_u8(f, 0x03)

@utils.register_class
class StrobeEffect(TimeWarp):
    class_id = b'STRB'
    propertydefs_dict = {}
    propertydefs = TimeWarp.propertydefs + [
        AVBPropertyDef('strobe_value',  'OMFI:STRB:MC:StrobeVal',  'int32'),
    ]
    __slots__ = ()

    def read(self, f):
        super(StrobeEffect, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)
        self.strobe_value = ctx.read_s32(f)
        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(StrobeEffect, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)
        ctx.write_s32(f, self.strobe_value)
        ctx.write_u8(f, 0x03)

@utils.register_class
class MotionEffect(TimeWarp):
    class_id = b'SPED'
    propertydefs_dict = {}
    propertydefs = TimeWarp.propertydefs + [
        AVBPropertyDef('speed_ratio',            'OMFI:SPED:Rate',                 'rational'),
        AVBPropertyDef('offset_adjust',          'OMIF:SPED:OffsetAdjust',         'double'),
        AVBPropertyDef('source_param_list',      'OMFI:SPED:SourceParamList',      'reference'),
        AVBPropertyDef('new_source_calculation', 'OMIF:SPED:NewSourceCalculation', 'bool'),
    ]
    __slots__ = ()

    def read(self, f):
        super(MotionEffect, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x03)

        num = ctx.read_s32(f)
        den = ctx.read_s32(f)
        self.speed_ratio = [num, den]

        for tag in ctx.iter_ext(f):

            if tag == 0x01:
                ctx.read_assert_tag(f, 75)
                self.offset_adjust = ctx.read_double(f)
            elif tag == 0x02:
                ctx.read_assert_tag(f, 72)
                self.source_param_list = ctx.read_object_ref(self.root, f)
            elif tag == 0x03:
                ctx.read_assert_tag(f, 66)
                self.new_source_calculation = ctx.read_bool(f)
            else:
                raise ValueError("%s: unknown ext tag 0x%02X %d" % (str(self.class_id), tag,tag))

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(MotionEffect, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x03)

        ctx.write_s32(f, self.speed_ratio[0])
        ctx.write_s32(f, self.speed_ratio[1])

        if hasattr(self, 'offset_adjust'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 75)
            ctx.write_double(f, self.offset_adjust)
        if hasattr(self, 'source_param_list'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x02)
            ctx.write_u8(f, 72)
            ctx.write_object_ref(self.root, f, self.source_param_list)
        if hasattr(self, 'new_source_calculation'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x03)
            ctx.write_u8(f, 66)
            ctx.write_bool(f, self.new_source_calculation)

        ctx.write_u8(f, 0x03)

@utils.register_class
class Repeat(TimeWarp):
    class_id = b'REPT'
    propertydefs_dict = {}
    __slots__ = ()

    def read(self, f):
        super(Repeat, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(Repeat, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)

        ctx.write_u8(f, 0x03)

@utils.register_class
class EssenceGroup(TrackGroup):
    class_id = b'RSET'
    propertydefs_dict = {}
    propertydefs = TrackGroup.propertydefs + [
        AVBPropertyDef('rep_set_type', 'OMFI:RSET:repSetType', 'int32'),
    ]
    __slots__ = ()

    def read(self, f):
        super(EssenceGroup, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        for tag in ctx.iter_ext(f):
            if tag == 0x01:
                ctx.read_assert_tag(f, 71)
                self.rep_set_type = ctx.read_s32(f)
            else:
                raise ValueError("%s: unknown ext tag 0x%02X %d" % (str(self.class_id), tag,tag))

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(EssenceGroup, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)

        if hasattr(self, 'rep_set_type'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.rep_set_type)

        ctx.write_u8(f, 0x03)

@utils.register_class
class TransitionEffect(TrackGroup):
    class_id = b'TNFX'
    propertydefs_dict = {}
    propertydefs = TrackGroup.propertydefs + [
        AVBPropertyDef('cutpoint',            'OMFI:TRAN:CutPoint',                   'int32'),
        AVBPropertyDef('left_length',         'OMFI:TNFX:MC:LeftLength',              'int32'),
        AVBPropertyDef('right_length',        'OMFI:TNFX:MC:RightLength',             'int32'),
        AVBPropertyDef('info_version',        'OMFI:TNFX:MC:GlobalInfoVersion',       'int16'),
        AVBPropertyDef('info_current',        'OMFI:TNFX:MC:GlobalInfo.kfCurrent',    'int32'),
        AVBPropertyDef('info_smooth',         'OMFI:TNFX:MC:GlobalInfo.kfSmooth',     'int32'),
        AVBPropertyDef('info_color_item',     'OMFI:TNFX:MC:GlobalInfo.colorItem',    'int16'),
        AVBPropertyDef('info_quality',        'OMFI:TNFX:MC:GlobalInfo.quality',      'int16'),
        AVBPropertyDef('info_is_reversed',    'OMFI:TNFX:MC:GlobalInfo.isReversed',   'int8'),
        AVBPropertyDef('info_aspect_on',      'OMFI:TNTNFXFX:MC:GlobalInfo.aspectOn', 'bool'),
        AVBPropertyDef('keyframes',           'OMFI:TNFX:MC:KeyFrameList',       'reference'),
        AVBPropertyDef('info_force_software', 'OMFI:TNFX:MC:ForceSoftware',           'bool'),
        AVBPropertyDef('info_never_hardware', 'OMFI:TNFX:MC:NeverHardware',           'bool'),
        AVBPropertyDef('trackman',            'OMFI:TNFX:MC:TrackMan',           'reference'),
    ]
    __slots__ = ()

    def read(self, f):
        super(TransitionEffect, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        self.cutpoint = ctx.read_s32(f)

        # the rest is the same as TKFX
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x05)

        self.left_length = ctx.read_s32(f)
        self.right_length = ctx.read_s32(f)

        self.info_version = ctx.read_s16(f)
        self.info_current = ctx.read_s32(f)
        self.info_smooth = ctx.read_s32(f)
        self.info_color_item = ctx.read_s16(f)
        self.info_quality = ctx.read_s16(f)
        self.info_is_reversed = ctx.read_s8(f)
        self.info_aspect_on = ctx.read_bool(f)

        self.keyframes = ctx.read_object_ref(self.root, f)
        self.info_force_software = ctx.read_bool(f)
        self.info_never_hardware = ctx.read_bool(f)

        for tag in ctx.iter_ext(f):
            if tag == 0x01:
                ctx.read_assert_tag(f, 72)
                self.trackman = ctx.read_object_ref(self.root, f)
            else:
                raise ValueError("%s: unknown ext tag 0x%02X %d" % (str(self.class_id), tag,tag))

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(TransitionEffect, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)

        ctx.write_s32(f, self.cutpoint)

        # the rest is the same as TKFX
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x05)

        ctx.write_s32(f, self.left_length)
        ctx.write_s32(f, self.right_length)

        ctx.write_s16(f, self.info_version)
        ctx.write_s32(f, self.info_current)
        ctx.write_s32(f, self.info_smooth)
        ctx.write_s16(f, self.info_color_item)
        ctx.write_s16(f, self.info_quality)
        ctx.write_s8(f, self.info_is_reversed)
        ctx.write_bool(f, self.info_aspect_on)

        ctx.write_object_ref(self.root, f, self.keyframes)
        ctx.write_bool(f, self.info_force_software)
        ctx.write_bool(f, self.info_never_hardware)

        if hasattr(self, 'trackman'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 72)
            ctx.write_object_ref(self.root, f, self.trackman)

        ctx.write_u8(f, 0x03)

@utils.register_class
class Selector(TrackGroup):
    class_id = b'SLCT'
    propertydefs_dict = {}
    propertydefs = TrackGroup.propertydefs + [
        AVBPropertyDef('is_ganged', 'OMFI:SLCT:IsGanged',      'bool'),
        AVBPropertyDef('selected',  'OMFI:SLCT:SelectedTrack', 'int16'),
    ]
    __slots__ = ()

    def read(self, f):
        super(Selector, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        self.is_ganged = ctx.read_bool(f)
        self.selected = ctx.read_u16(f)

        assert self.selected < len(self.tracks)

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(Selector, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)

        ctx.write_bool(f, self.is_ganged)
        ctx.write_u16(f, self.selected)

        ctx.write_u8(f, 0x03)

    def components(self):
        for track in self.tracks:
            yield track.component

@utils.register_class
class Composition(TrackGroup):
    class_id = b'CMPO'
    propertydefs_dict = {}
    propertydefs = TrackGroup.propertydefs + [
        AVBPropertyDef('last_modified', 'OMFI:MOBJ:LastModified',  'int32',        0),
        AVBPropertyDef('mob_type_id',   '__OMFI:MOBJ:MobType',     'int8',         2),
        AVBPropertyDef('usage_code',    'OMFI:MOBJ:UsageCode',     'int8',         0),
        AVBPropertyDef('descriptor',    'OMFI:MOBJ:PhysicalMedia', 'reference', None),
        AVBPropertyDef('creation_time', 'OMFI:MOBJ:_CreationTime', 'int32'),
        AVBPropertyDef('mob_id',        'MobID',                   'MobID'),
    ]
    __slots__ = ()

    def __init__(self, name='Mob', mob_type="MasterMob"):
        super(Composition, self).__init__()

        self.name = mob_type
        if mob_type == "CompositionMob":
            self.mob_type_id = 1

        elif mob_type == "MasterMob":
            self.mob_type_id = 2

        elif mob_type == "SourceMob":
            self.mob_type_id = 3
        else:
            ValueError("Unknown mob type: %d" % mob_type)

        self.mob_id = mobid.MobID.new()
        now = datetime.datetime.now()
        self.last_modified = now
        self.creation_time = now

    def read(self, f):
        super(Composition, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x02)

        mob_id_lo = ctx.read_u32(f)
        mob_id_hi = ctx.read_u32(f)
        self.last_modified = ctx.read_datetime(f)

        self.mob_type_id = ctx.read_u8(f)
        self.usage_code =  ctx.read_s32(f)
        self.descriptor = ctx.read_object_ref(self.root, f)

        for tag in ctx.iter_ext(f):

            if tag == 0x01:
                ctx.read_assert_tag(f, 71)
                self.creation_time = ctx.read_datetime(f)
            elif tag == 0x02:
                self.mob_id = ctx.read_mob_id(f)
            else:
                raise ValueError("%s: unknown ext tag 0x%02X %d" % (str(self.class_id), tag,tag))

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(Composition, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x02)

        lo = self.mob_id.material.time_low
        hi = self.mob_id.material.time_mid + (self.mob_id.material.time_hi_version << 16)
        ctx.write_u32(f, lo)
        ctx.write_u32(f, hi)
        ctx.write_datetime(f, self.last_modified)

        ctx.write_u8(f, self.mob_type_id)
        ctx.write_s32(f, self.usage_code)
        ctx.write_object_ref(self.root, f, self.descriptor)

        if hasattr(self, 'creation_time'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 71)
            ctx.write_datetime(f, self.creation_time)

        if hasattr(self, 'mob_id'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x02)
            ctx.write_mob_id(f, self.mob_id)

        ctx.write_u8(f, 0x03)

    @property
    def usage(self):
        # these usage codes seem to come from omf
        if self.usage_code == 0:
            return None
        # master mob to a precompute
        elif self.usage_code == 1:
            return "precompute"
        # mob is a subclip
        elif self.usage_code == 2:
            return "subclip"
        # mob is an effect holder
        elif self.usage_code == 3:
            return "effect"
        # comp of selectors
        elif self.usage_code == 4:
            return "group"
        # mob that back up groups
        elif self.usage_code == 5:
            return "groupoofter"
        # motion effect clip
        elif self.usage_code == 6:
            return "motion"
        # group phys mobs
        elif self.usage_code == 7:
            return "mastermob"
        # file mob with a precompute
        elif self.usage_code == 9:
            return "precompute_file"
        else:
            return "unknown"

        #TODO: find out what code  8,10 -> 14 are
        # 14 essencegroup?

    @property
    def mob_type(self):
        if self.mob_type_id == 1:
            return "CompositionMob"
        elif self.mob_type_id == 2:
            return "MasterMob"
        elif self.mob_type_id == 3:
            return "SourceMob"
        else:
            raise ValueError("Unknown mob type id: %d" % self.mob_type_id)



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )

import datetime

import struct
import io
import os
import binascii
import traceback
import array
from weakref import WeakValueDictionary
import struct

from . import utils
from .core import walk_references
from .ioctx import AVBIOContext


try:
    from ._ext import READERS
except:
    READERS = {}

class AVBChunk(object):
    __slots__ = ('root', 'class_id', 'pos', 'size')
    def __init__(self, root, class_id, pos, size):
        self.root = root
        self.class_id = class_id
        self.pos = pos
        self.size = size

    def read(self):
        self.root.f.seek(self.pos)
        return self.root.f.read(self.size)

    def hex(self):
        ctx = self.root.ictx

        header = ctx.reverse_str(self.class_id)
        size = struct.pack(b"<I", self.size)
        data =  header + size + self.read()
        return binascii.hexlify(data)

def read_chunk(root, f):
    ctx = root.ictx
    class_id = ctx.read_fourcc(f)
    size = ctx.read_u32(f)
    pos = f.tell()

    return AVBChunk(root, class_id, pos, size)

class AVBFactory(object):

    def __init__(self, root):
        self.root = root
        self.class_name = None

    def __getattr__(self, name):
        self.class_name = name
        return self.create_instance

    def from_name(self, name, *args, **kwargs):

        classobj = obj_class = utils.AVBClassName_dict.get(name, None)

        # obj = classobj(None, *args, **kwargs)
        obj = classobj.__new__(classobj)
        obj.root = self.root
        if obj.class_id:
            self.root.next_object_id += 1
            obj.instance_id = self.root.next_object_id

            self.root.modified_objects[obj.instance_id] = obj
            self.root.object_cache[obj.instance_id] = obj

        obj.__init__(*args, **kwargs)

        return obj

    def create_instance(self, *args, **kwargs):
        return self.from_name(self.class_name, *args, **kwargs)

LE_BYTE_ORDER = b'\x06\x00'
BE_BYTE_ORDER = b'\x00\x06'
MAGIC=b'Domain'

def is_fileobject_like(fileobject):

    for attr_name in ('read', 'readinto', 'seek', 'tell', 'close'):
        if not hasattr(fileobject, attr_name):
            return False

    return True

class AVBFile(object):
    def __init__(self, fileobject=None, buffering=io.DEFAULT_BUFFER_SIZE, use_ext=True):

        self.check_refs = True
        self.debug_copy_refs = False
        self.reading = False

        self.create = AVBFactory(self)
        self.object_cache = WeakValueDictionary()
        self.modified_objects = {}
        self.next_object_id = 0

        if fileobject is None:
            self.setup_empty()
            return

        if is_fileobject_like(fileobject):
            self.f = fileobject
        else:
            self.f = io.open(fileobject, 'rb', buffering=buffering)

        f = self.f
        file_bytes = f.read(2)
        self.fast_readers = {}
        if file_bytes == LE_BYTE_ORDER:
            ctx = AVBIOContext('little')
            if use_ext:
                self.fast_readers = READERS
        elif file_bytes == BE_BYTE_ORDER:
            ctx = AVBIOContext('big')
        else:
            raise ValueError("not a avb file")

        self.ictx = ctx

        header = f.read(len(MAGIC))
        if header != MAGIC:
            raise ValueError("not a avb file")

        pos = f.tell()

        assert ctx.read_fourcc(f) == b'OBJD'
        assert ctx.read_string(f) == u'AObjDoc'
        assert ctx.read_u8(f) == 0x04

        last_save_str = ctx.read_string(f)
        num_objects = ctx.read_u32(f)
        self.root_index = ctx.read_u32(f)

        v  = ctx.read_u32(f)
        assert v in (0x49494949, 0x4D4D4D4D)

        self.last_save = ctx.read_datetime(f)

        # skip 4 bytes
        f.read(4)

        assert ctx.read_fourcc(f) == b'ATob'
        assert ctx.read_fourcc(f) == b'ATve'

        self.creator_version = ctx.read_string(f)

        # Reserved data
        f.read(16)

        self.root_chunk = AVBChunk(self, b'OBJD', pos, f.tell() - pos)

        self.object_positions = array.array(str('L'), [0 for i in range(num_objects+1)])
        self.next_object_id = len(self.object_positions)

        for i in range(num_objects):
            self.object_positions[i+1] = f.tell()
            class_id = f.read(4)
            size = ctx.read_u32(f)

            f.seek(size, os.SEEK_CUR)

        self.content = self.read_object(self.root_index)

    def setup_empty(self):
        self.f = None

        self.creator_version = u'pyavb 0.1.0'
        self.content = self.create.Bin()

        self.update_save_time()

    def update_save_time(self):
        self.last_save = datetime.datetime.now()

    def add_modified(self, obj):
        self.modified_objects[obj.instance_id] = obj

    def write_header(self, f):

        ctx = self.octx

        if ctx.byte_order == 'little':
            f.write(LE_BYTE_ORDER)
        else:
            f.write(BE_BYTE_ORDER)

        f.write(MAGIC)
        ctx.write_fourcc(f, b'OBJD')
        ctx.write_string(f, u'AObjDoc')
        ctx.write_u8(f, 0x04)
        last_save_str = self.last_save.strftime(u'%Y/%m/%d %H:%M:%S')
        ctx.write_string(f, last_save_str)
        pos = f.tell()
        ctx.write_u32(f, 0)
        ctx.write_u32(f, 0)

        if ctx.byte_order == 'little':
            ctx.write_u32(f, 0x49494949)
        else:
            ctx.write_u32(f, 0x4D4D4D4D)

        ctx.write_datetime(f, self.last_save)
        ctx.write_u32(f, 0)

        ctx.write_fourcc(f, b'ATob')
        ctx.write_fourcc(f, b'ATve')

        # version =

        s = f.tell()
        v = self.creator_version.encode('macroman')
        v = v[:30]
        ctx.write_u16(f, 30)
        f.write(v)

        # pad with 0x20
        while f.tell() - s < 32:
            ctx.write_u8(f, 0x20)
        f.write(bytearray(16))

        return pos

    def read_chunk(self, index):
        if index == 0:
            return self.root_chunk

        object_pos = self.object_positions[index]

        f = self.f
        f.seek(object_pos)

        if self.ictx.byte_order == 'little':
            class_id, size = struct.unpack(b"<4sI", f.read(8))
            class_id = class_id[::-1]
        else:
            class_id, size = struct.unpack(b">4sI", f.read(8))

        pos = f.tell()
        chunk = AVBChunk(self, class_id, pos, size)
        return chunk

    def read_object(self, index):
        if index == 0:
            return None

        object_instance = self.object_cache.get(index, None)
        if object_instance is not None:
            return object_instance

        object_pos = self.object_positions[index]

        f = self.f
        f.seek(object_pos)

        if self.ictx.byte_order == 'little':
            class_id, size = struct.unpack(b"<4sI", f.read(8))
            class_id = class_id[::-1]
        else:
            class_id, size = struct.unpack(b">4sI", f.read(8))

        data = bytearray(size)
        bytes_read = f.readinto(data)
        assert bytes_read == size

        obj_class = utils.AVBClaseID_dict.get(class_id, None)
        if obj_class:
            try:
                self.reading = True
                # NOTE: objects read from file do not run __init__
                object_instance = obj_class.__new__(obj_class, root=self)

                # Only OrderedDict needs run __init__ in order to work
                if class_id == b'ATTR':
                    object_instance.__init__(object_instance)

                reader = self.fast_readers.get(class_id, None)

                if reader:
                    reader(self, object_instance, data)
                else:
                    r = io.BytesIO(data)
                    object_instance.read(r)
                    # print(len(r.read()))
                    assert len(r.read()) == 0
                self.object_cache[index] = object_instance
                object_instance.instance_id = index
                return object_instance
            except:
                pos = self.object_positions[index] + 8
                chunk = AVBChunk(self, class_id, pos, len(data))
                print(chunk.class_id)
                print(chunk.hex())
                print(traceback.format_exc())
                raise
            finally:
                self.reading = False

        else:
            pos = self.object_positions[index] + 8
            chunk = AVBChunk(self, class_id, pos, len(data))
            print(chunk.class_id)
            print(chunk.hex())
            raise NotImplementedError(chunk.class_id)

    def write_object(self, f, obj):
        buffer = io.BytesIO()
        obj.write(buffer)
        data = buffer.getvalue()
        assert data[-1:] == b'\x03'
        ctx = self.octx

        ctx.write_fourcc(f, obj.class_id)
        ctx.write_u32(f, len(data))
        f.write(data)

        # chunk = self.read_chunk(obj.instance_id)
        # orig_chunk_data = chunk.read()

        # if len(orig_chunk_data) != len(data):# or orig_chunk_data != data:
        #     print(obj, len(orig_chunk_data), len(data) )
        #     print(binascii.hexlify(orig_chunk_data))
        #     print("")
        #     print(binascii.hexlify(data))
        #     raise Exception()

    def write(self, path, byte_order='little'):
        self.next_chunk_id = 0
        self.ref_mapping = {}
        ctx = AVBIOContext(byte_order)
        self.octx = ctx

        with io.open(path, 'wb') as f:
            count_pos = self.write_header(f)
            for obj in walk_references(self.content):
                if obj.instance_id in self.ref_mapping:
                    continue

                self.next_chunk_id += 1
                self.ref_mapping[obj.instance_id] = self.next_chunk_id
                self.write_object(f, obj)

            pos = f.tell()
            f.seek(count_pos)
            ctx.write_u32(f, self.next_chunk_id)
            ctx.write_u32(f, self.next_chunk_id)
            f.seek(pos)

    def chunks(self):
        for i in range(len(self.object_positions)):
            yield self.read_chunk(i)

    def iter_class_ids(self, class_id_list):
        for i, chunk in enumerate(self.chunks()):
            if chunk.class_id in class_id_list:
                yield self.read_object(i)

    def close(self):
        if self.f:
            self.f.close()

    def __exit__(self, exc_type, exc_value, traceback):
        if (exc_type is None and exc_value is None and traceback is None):
            self.close()

    def __enter__(self):
        return self



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )

import time
from datetime import datetime
from struct import (pack, unpack)
from uuid import UUID

from .utils import AVBObjectRef
from .mobid import MobID

exp10_pretty = {
5994: (59940, -3),
}

class AVBIOContext(object):
    def __init__(self, byte_order='little'):
        self.byte_order = byte_order

        if byte_order == 'little':
            self.read_u16  = self.read_u16le
            self.write_u16 = self.write_u16le
            self.read_s16  = self.read_s16le
            self.write_s16 = self.write_s16le
            self.read_u32  = self.read_u32le
            self.write_u32 = self.write_u32le
            self.read_s32  = self.read_s32le
            self.write_s32 = self.write_s32le
            self.read_u64  = self.read_u64le
            self.write_u64 = self.write_u64le
            self.read_s64  = self.read_s64le
            self.write_s64 = self.write_s64le

            self.read_double  = self.read_double_le
            self.write_double = self.write_double_le

            self.read_fourcc    = self.read_fourcc_le
            self.write_fourcc   = self.write_fourcc_le

        elif byte_order == 'big':
            self.read_u16  = self.read_u16be
            self.write_u16 = self.write_u16be
            self.read_s16  = self.read_s16be
            self.write_s16 = self.write_s16be
            self.read_u32  = self.read_u32be
            self.write_u32 = self.write_u32be
            self.read_s32  = self.read_s32be
            self.write_s32 = self.write_s32be
            self.read_u64  = self.read_u64be
            self.write_u64 = self.write_u64be
            self.read_s64  = self.read_s64be
            self.write_s64 = self.write_s64be

            self.read_double = self.read_double_be
            self.write_double = self.write_double_be

            self.read_fourcc    = self.read_fourcc_be
            self.write_fourcc   = self.write_fourcc_be
        else:
            raise ValueError('bytes_order must be "big" or "little"')

    @staticmethod
    def read_assert_tag(f, version):
        version_mark = AVBIOContext.read_u8(f)
        if version_mark != version:
            raise AssertionError("%d != %d" % (version_mark, version))

    @staticmethod
    def iter_ext(f):
        while True:
            pos = f.tell()
            tag = AVBIOContext.read_u8(f)
            if tag != 0x01:
                f.seek(pos)
                break

            tag = AVBIOContext.read_u8(f)
            yield tag

    @staticmethod
    def reverse_str(s):
        size = len(s)
        result = bytearray(size)
        for i in range(size):
            result[size - 1 - i] = s[i]

        return bytes(result)

    @staticmethod
    def datetime_to_timestamp(d):
        return int(time.mktime(d.timetuple()))

    @staticmethod
    def read_u8(f):
        (result, ) = unpack(b"B", f.read(1))
        return result

    @staticmethod
    def write_u8(f, value):
        f.write(pack(b"B", value))

    @staticmethod
    def read_s8(f):
        (result, ) = unpack(b"b", f.read(1))
        return result

    @staticmethod
    def write_s8(f, value):
        f.write(pack(b"b", value))

    @staticmethod
    def read_bool(f):
        return AVBIOContext.read_u8(f) == 0x01

    @staticmethod
    def write_bool(f, value):
        if value:
            AVBIOContext.write_u8(f, 0x01)
        else:
            AVBIOContext.write_u8(f, 0x00)

    # complex data types

    def read_exp10_encoded_float(self, f):
        mantissa = self.read_s32(f)
        exp10 = self.read_s16(f)

        return float(mantissa) * pow(10, exp10)

    def write_exp10_encoded_float(self, f, value):
        exponent = 0
        while int(value) != value:
            if abs(value * 10) >= 0x7FFFFFFF:
                break
            if exponent <= -6:
                break
            value *= 10
            exponent -= 1

        # remap values pretty values to match seen files
        if value in exp10_pretty:
            self.write_s32(f, exp10_pretty[value][0])
            self.write_s16(f, exp10_pretty[value][1])
        else:
            self.write_s32(f, int(value))
            self.write_s16(f, exponent)


    def read_string(self, f, encoding = 'macroman'):
        size = self.read_u16(f)
        if size >= 65535:
            return u""

        s = f.read(size)
        s = s.strip(b'\x00')
        return s.decode(encoding)

    def write_string(self, f, s, encoding = 'macroman'):
        s = s or b""
        if s == b"":
            self.write_u16(f, 0)
            return

        data = s.encode(encoding)
        if encoding == 'utf-8':
            data = b'\x00\x00' + data

        size = len(data)
        self.write_u16(f, size)
        f.write(data)

    def read_object_ref(self, root, f):
        index = self.read_u32(f)
        ref =  AVBObjectRef(root, index)
        if not root.check_refs or ref.valid:
            return ref
        raise ValueError("bad index: %d" % index)

    def write_object_ref(self, root, f, value):
        if value is None:
            index = 0
        elif root.debug_copy_refs:
            index = value.index
        elif value.instance_id not in root.ref_mapping:
            raise Exception("object not written yet")
        else:
            index = root.ref_mapping[value.instance_id]

        self.write_u32(f, index)

    def read_uuid(self, f):
        data = b''
        self.read_assert_tag(f, 72)
        data += f.read(4)

        self.read_assert_tag(f, 70)
        data += f.read(2)

        self.read_assert_tag(f, 70)
        data += f.read(2)

        self.read_assert_tag(f, 65)
        data4len = self.read_s32(f)
        assert data4len == 8
        data += f.read(8)

        if self.byte_order == 'little':
            return UUID(bytes_le=data)
        else:
            return UUID(bytes=data)

    def write_uuid(self, f, value):
        self.write_u8(f, 72)
        self.write_u32(f, value.time_low)
        self.write_u8(f, 70)
        self.write_u16(f, value.time_mid)
        self.write_u8(f, 70)
        self.write_u16(f, value.time_hi_version)

        self.write_u8(f, 65)
        self.write_s32(f, 8)
        if self.byte_order == 'little':
            f.write(value.bytes_le[8:])
        else:
            f.write(value.bytes[8:])

    def read_mob_id(self, f):
        m = MobID()
        self.read_assert_tag(f, 65)
        smpte_label_len = self.read_s32(f)
        assert smpte_label_len == 12

        m.SMPTELabel = [self.read_u8(f) for i in range(12)]

        self.read_assert_tag(f, 68)
        m.length = self.read_u8(f)

        self.read_assert_tag(f, 68)
        m.instanceHigh = self.read_u8(f)

        self.read_assert_tag(f, 68)
        m.instanceMid = self.read_u8(f)

        self.read_assert_tag(f, 68)
        m.instanceLow = self.read_u8(f)

        m.material = self.read_uuid(f)
        return m

    def write_mob_id(self, f, m):

        self.write_u8(f, 65)
        self.write_s32(f, 12)
        for i in m.SMPTELabel:
            self.write_u8(f, i)

        self.write_u8(f, 68)
        self.write_u8(f, m.length)

        self.write_u8(f, 68)
        self.write_u8(f, m.instanceHigh)

        self.write_u8(f, 68)
        self.write_u8(f, m.instanceMid)

        self.write_u8(f, 68)
        self.write_u8(f, m.instanceLow)

        self.write_uuid(f, m.material)

    def read_raw_uuid(self, f):
        if self.byte_order == 'little':
            return UUID(bytes_le=f.read(16))
        else:
            return UUID(bytes=f.read(16))

    def write_raw_uuid(self, f, value):
        if self.byte_order == 'little':
            f.write(value.bytes_le)
        else:
            f.write(value.bytes)

    def read_datetime(self, f):
        return datetime.fromtimestamp(self.read_u32(f))

    def write_datetime(self, f, value):
        self.write_u32(f, self.datetime_to_timestamp(value))

    def read_rect(self, f):
        version = self.read_s16(f)
        assert version == 1

        a = self.read_s16(f)
        b = self.read_s16(f)
        c = self.read_s16(f)
        d = self.read_s16(f)

        return [a,b,c,d]

    def write_rect(self, f, v):
        self.write_s16(f, 1)
        self.write_s16(f, v[0])
        self.write_s16(f, v[1])
        self.write_s16(f, v[2])
        self.write_s16(f, v[3])

    def read_rgb_color(self, f):
        version = self.read_s16(f)
        assert version == 1
        r = self.read_u16(f)
        g = self.read_u16(f)
        b = self.read_u16(f)

        return [r,g,b]

    def write_rgb_color(self, f, v):
        self.write_s16(f, 1)
        self.write_u16(f, v[0])
        self.write_u16(f, v[1])
        self.write_u16(f, v[2])

    # little

    @staticmethod
    def read_u16le(f):
        (result, ) = unpack(b"<H", f.read(2))
        return result

    @staticmethod
    def write_u16le(f, value):
        f.write(pack(b"<H", value))

    @staticmethod
    def read_s16le(f):
        (result, ) = unpack(b"<h", f.read(2))
        return result

    @staticmethod
    def write_s16le(f, value):
        f.write(pack(b"<h", value))

    @staticmethod
    def read_u32le(f):
        (result, ) = unpack(b"<I", f.read(4))
        return result

    @staticmethod
    def write_u32le(f, value):
        return f.write(pack(b"<I", value))

    @staticmethod
    def read_s32le(f):
        (result, ) = unpack(b"<i", f.read(4))
        return result

    @staticmethod
    def write_s32le(f, value):
        return f.write(pack(b"<i", value))

    @staticmethod
    def read_u64le(f):
        return unpack(b"<Q", f.read(8))[0]

    @staticmethod
    def write_u64le(f, value):
        return f.write(pack(b"<Q", value))

    @staticmethod
    def read_s64le(f):
        return unpack(b"<q", f.read(8))[0]

    @staticmethod
    def write_s64le(f, value):
        return f.write(pack(b"<q", value))

    @staticmethod
    def read_double_le(f):
        return unpack(b"<d", f.read(8))[0]

    @staticmethod
    def write_double_le(f, value):
        f.write(pack(b"<d", value))

    @staticmethod
    def read_fourcc_le(f):
        return f.read(4)[::-1]

    @staticmethod
    def write_fourcc_le(f, value):
        assert len(value) == 4
        f.write(AVBIOContext.reverse_str(value))

    # big

    @staticmethod
    def read_u16be(f):
        (result, ) = unpack(b">H", f.read(2))
        return result

    @staticmethod
    def write_u16be(f, value):
        f.write(pack(b">H", value))

    @staticmethod
    def read_s16be(f):
        (result, ) = unpack(b">h", f.read(2))
        return result

    @staticmethod
    def write_s16be(f, value):
        f.write(pack(b">h", value))

    @staticmethod
    def read_u32be(f):
        (result, ) = unpack(b">I", f.read(4))
        return result

    @staticmethod
    def write_u32be(f, value):
        return f.write(pack(b">I", value))

    @staticmethod
    def read_s32be(f):
        (result, ) = unpack(b">i", f.read(4))
        return result

    @staticmethod
    def write_s32be(f, value):
        return f.write(pack(b">i", value))

    @staticmethod
    def read_u64be(f):
        return unpack(b">Q", f.read(8))[0]

    @staticmethod
    def write_u64be(f, value):
        return f.write(pack(b">Q", value))

    @staticmethod
    def read_s64be(f):
        return unpack(b">q", f.read(8))[0]

    @staticmethod
    def write_s64be(f, value):
        return f.write(pack(b">q", value))

    @staticmethod
    def read_double_be(f):
        return unpack(b"<d", f.read(8))[0]

    @staticmethod
    def write_double_be(f, value):
        f.write(pack(b"<d", value))

    @staticmethod
    def read_fourcc_be(f):
        return f.read(4)

    @staticmethod
    def write_fourcc_be(f, value):
        assert len(value) == 4
        f.write(value)



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )

from . import core
from . import utils
from . core import AVBPropertyDef
from . utils import peek_data

class Setting(core.AVBObject):
    class_id = b'ASET'
    propertydefs_dict = {}
    propertydefs = [
        AVBPropertyDef('name',       'name',       'string'),
        AVBPropertyDef('kind',       'kind',       'string'),
        AVBPropertyDef('attr_count', 'attributes', 'int16',      1),
        AVBPropertyDef('attr_type',  'type',       'int16',     20),
        AVBPropertyDef('attributes', 'attrList',   'reference'),
    ]
    __slots__ = ()

    def read(self, f):
        super(Setting, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x06)

        self.name = ctx.read_string(f)
        self.kind = ctx.read_string(f)

        self.attr_count = ctx.read_s16(f)
        self.attr_type =  ctx.read_s16(f)
        self.attributes = ctx.read_object_ref(self.root, f)

    def write(self, f):
        super(Setting, self).write(f)

        ctx = self.root.octx

        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x06)

        ctx.write_string(f, self.name)
        ctx.write_string(f, self.kind)

        ctx.write_s16(f, self.attr_count)
        ctx.write_s16(f, self.attr_type)
        ctx.write_object_ref(self.root, f, self.attributes)

default_bin_columns = [
    {u'title': u'Color',         u'format': 105,  u'type': 51,  u'hidden': False},
    {u'title': u'   ',           u'format': 0,    u'type': 200, u'hidden': False},
    {u'title': u'Name',          u'format': 20,   u'type': 201, u'hidden': False},
    {u'title': u'Creation Date', u'format': 102,  u'type': 12,  u'hidden': False},
    {u'title': u'Duration',      u'format': 100,  u'type': 4,   u'hidden': False},
    {u'title': u'Drive',         u'format': 2,    u'type': 14,  u'hidden': False},
    {u'title': u'IN-OUT',        u'format': 100,  u'type': 7,   u'hidden': False},
    {u'title': u'Mark IN',       u'format': 100,  u'type': 5,   u'hidden': False},
    {u'title': u'Mark OUT',      u'format': 100,  u'type': 6,   u'hidden': False},
    {u'title': u'Tracks',        u'format': 2,    u'type': 1,   u'hidden': False},
    {u'title': u'Start',         u'format': 100,  u'type': 2,   u'hidden': False},
    {u'title': u'Tape',          u'format': 2,    u'type': 8,   u'hidden': False},
    {u'title': u'Video',         u'format': 2,    u'type': 13,  u'hidden': False},
    {u'title': u'Plug-in',       u'format': 2,    u'type': 133, u'hidden': False},
    {u'title': u'TapeID',        u'format': 2,    u'type': 50,  u'hidden': False},
    {u'title': u'Audio SR',      u'format': 2,    u'type': 11,  u'hidden': False},
    {u'title': u'Comments',      u'format': 2,    u'type': 40,  u'hidden': False},
]

@utils.register_class
class BinViewSetting(Setting):
    class_id = b'BVst'
    propertydefs_dict = {}
    propertydefs = Setting.propertydefs + [
        AVBPropertyDef('columns',              'Columns',          'list'),
        AVBPropertyDef('format_descriptors',   'FormatDescriptor', 'list'),
    ]
    __slots__ = ()

    def __init__(self):
        super(BinViewSetting, self).__init__(self)
        self.name = u'Untitled'
        self.kind = u'Bin View'
        self.attributes = self.root.create.Attributes()
        self.format_descriptors = []
        self.columns = []
        self.columns.extend(default_bin_columns)

    def read(self, f):
        super(BinViewSetting, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 10)

        self.columns = []

        column_count = ctx.read_u16(f)
        for i in range(column_count):
            d = {}
            d['title']  = ctx.read_string(f)
            d['format'] = ctx.read_s16(f)
            d['type']   = ctx.read_s16(f)
            d['hidden'] = ctx.read_bool(f)
            self.columns.append(d)

        for tag in ctx.iter_ext(f):
            if tag == 0x01:
                ctx.read_assert_tag(f, 69)
                num_vcid_free_columns = ctx.read_s16(f)
                assert num_vcid_free_columns >= 0
                self.format_descriptors = []
                for i in range(num_vcid_free_columns):
                    d = {}

                    ctx.read_assert_tag(f, 69)

                    # vcid == Video Codec ID?
                    d['vcid_free_column_id'] = ctx.read_s16(f)

                    ctx.read_assert_tag(f, 71)
                    format_descriptor_size = ctx.read_s32(f)

                    ctx.read_assert_tag(f, 76)

                    # utf-8 seems to start with 4 null bytes
                    ctx.read_s32(f)
                    d['format_descriptor'] = f.read(format_descriptor_size).decode('utf8')
                    self.format_descriptors.append(d)
            else:
                raise ValueError("%s: unknown ext tag 0x%02X %d" % (str(self.class_id), tag,tag))

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(BinViewSetting, self).write(f)
        ctx = self.root.octx

        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 10)
        ctx.write_u16(f, len(self.columns))

        for d in self.columns:
            ctx.write_string(f, d['title'])
            ctx.write_s16(f, d['format'])
            ctx.write_s16(f, d['type'])
            ctx.write_bool(f, d['hidden'])

        if hasattr(self, 'format_descriptors'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 69)
            ctx.write_u16(f, len(self.format_descriptors))

            for d in self.format_descriptors:
                ctx.write_u8(f, 69)
                ctx.write_s16(f, d['vcid_free_column_id'])

                format_data =  d['format_descriptor'].encode('utf-8')
                ctx.write_u8(f, 71)
                ctx.write_s32(f, len(format_data))

                ctx.write_u8(f, 76)
                #null bytes
                ctx.write_s16(f,  len(format_data) + 2)
                ctx.write_s16(f, 0)
                f.write(format_data)

        ctx.write_u8(f, 0x03)

@utils.register_helper_class
class BinItem(core.AVBObject):
    propertydefs_dict = {}
    propertydefs = [
        AVBPropertyDef('mob',         'Composition',  'reference'),
        AVBPropertyDef('x',           'Xpos',         'int16',    -30000),
        AVBPropertyDef('y',           'Ypos',         'int16',    -30000),
        AVBPropertyDef('keyframe',    'Keyframe',     'int32',         0),
        AVBPropertyDef('user_placed', 'userPlaced',   'bool',       True),
    ]
    __slots__ = ()

@utils.register_helper_class
class SiftItem(core.AVBObject):
    propertydefs_dict = {}
    propertydefs = [
        AVBPropertyDef('method', 'SiftMethod', 'int16' ,       1),
        AVBPropertyDef('string', 'SiftString', 'string',     u''),
        AVBPropertyDef('column', 'SiftColumn', 'string',  u'Any'),
    ]
    __slots__ = ()

@utils.register_class
class Bin(core.AVBObject):
    class_id = b'ABIN'
    propertydefs_dict = {}
    propertydefs = [
        AVBPropertyDef('large_bin',         'large_bin',      'bool',     False), #custom
        AVBPropertyDef('view_setting',   'binviewsetting', 'reference'),
        AVBPropertyDef('uid',            'binuid.high',    'uint64'),
        AVBPropertyDef('items',            'Items',          'list',           ), # custom
        AVBPropertyDef('display_mask',     'DisplayMask',    'int32',     98999),
        AVBPropertyDef('display_mode',     'DisplayMode',    'int32',         0),
        AVBPropertyDef('sifted',           'Sifted',         'bool',      False),
        AVBPropertyDef('sifted_settings',  'SiftedSettring', 'list'), #custom
        AVBPropertyDef('sort_columns',     'SortColumns',    'list',           ),
        AVBPropertyDef('mac_font',         'MacFont',        'int16',         1),
        AVBPropertyDef('mac_font_size',    'MacFontSize',    'int16',        11),
        AVBPropertyDef('mac_image_scale',  'MacImageScale',  'int16',         5),
        AVBPropertyDef('home_rect',        'HomeRect',       'rect',       [0, 0 , 300, 600]),
        AVBPropertyDef('background_color', 'BackColor',      'color', [45568, 45568, 45568]),
        AVBPropertyDef('forground_color',  'ForeColor',      'color',    [3328, 3328, 3328]),
        AVBPropertyDef('ql_image_scale',   'QLImageScale',   'int16',         6),
        AVBPropertyDef('attributes',       'BinAttr',        'reference'),
        AVBPropertyDef('was_iconic',       'WasIconic',      'bool',      False),
    ]
    __slots__ = ('mob_dict')

    def __init__(self):
        super(Bin, self).__init__(self)

        self.view_setting = self.root.create.BinViewSetting()
        self.uid = utils.generate_uid()
        self.items = []
        self.sort_columns = []

        self.sifted_settings= []
        for i in range(6):
            s = self.root.create.SiftItem()
            self.sifted_settings.append(s)

        self.attributes = self.root.create.Attributes()

    def read(self, f):
        super(Bin, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)

        version = ctx.read_u8(f)
        assert version in (0x0e, 0x0f)
        if version ==  0x0f:
            self.large_bin = True
        else:
            self.large_bin = False

        self.view_setting = ctx.read_object_ref(self.root, f)
        self.uid = ctx.read_u64(f)

        if version == 0x0e:
            object_count = ctx.read_u16(f)
        else:
            #large bin size > max u16
            object_count = ctx.read_u32(f)

        self.items = []

        for i in range(object_count):
            bin_obj = BinItem.__new__(BinItem, root=self.root)
            bin_obj.mob = ctx.read_object_ref(self.root, f)
            bin_obj.x = ctx.read_s16(f)
            bin_obj.y = ctx.read_s16(f)
            bin_obj.keyframe = ctx.read_s32(f)
            bin_obj.user_placed = ctx.read_bool(f)
            self.items.append(bin_obj)

        self.display_mask = ctx.read_s32(f)
        self.display_mode = ctx.read_s16(f)

        # sifted stuff for searching settings bin
        # don't care too much about it at the moment
        self.sifted = ctx.read_bool(f)
        self.sifted_settings= []

        for i in range(6):
            s = SiftItem.__new__(SiftItem, root=self.root)
            s.method = ctx.read_s16(f)
            s.string = ctx.read_string(f)
            s.column = ctx.read_string(f)
            self.sifted_settings.append(s)

        sort_column_count = ctx.read_s16(f)
        self.sort_columns = []
        for i in range(sort_column_count):
            direction = ctx.read_u8(f)
            col = ctx.read_string(f)
            self.sort_columns.append([direction, col])


        self.mac_font = ctx.read_s16(f)
        self.mac_font_size = ctx.read_s16(f)
        self.mac_image_scale = ctx.read_s16(f)

        self.home_rect = ctx.read_rect(f)

        self.background_color = ctx.read_rgb_color(f)
        self.forground_color = ctx.read_rgb_color(f)

        self.ql_image_scale = ctx.read_s16(f)

        self.attributes = ctx.read_object_ref(self.root, f)
        self.was_iconic = ctx.read_bool(f)

        if self.class_id[:] == b'ABIN':
            ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(Bin, self).write(f)
        ctx = self.root.octx

        ctx.write_u8(f, 0x02)

        object_count = len(self.items)
        #large bin size > max u16
        if self.large_bin or object_count > 0xffff:
            ctx.write_u8(f,  0x0f)
        else:
            ctx.write_u8(f,  0x0e)

        ctx.write_object_ref(self.root, f, self.view_setting)
        ctx.write_u64(f, self.uid)

        #large bin size > max u16
        if self.large_bin or object_count > 0xffff:
            ctx.write_u32(f, object_count)
        else:
            ctx.write_u16(f, object_count)

        for bin_obj in self.items:
             ctx.write_object_ref(self.root, f, bin_obj.mob)
             ctx.write_s16(f, bin_obj.x)
             ctx.write_s16(f, bin_obj.y)
             ctx.write_s32(f, bin_obj.keyframe)
             ctx.write_u8(f, bin_obj.user_placed)

        ctx.write_s32(f, self.display_mask)
        ctx.write_s16(f, self.display_mode)
        ctx.write_bool(f, self.sifted)

        for i in range(6):
            s = self.sifted_settings[i]
            ctx.write_s16(f, s.method)
            ctx.write_string(f, s.string)
            ctx.write_string(f, s.column)


        ctx.write_s16(f, len(self.sort_columns))

        for col in self.sort_columns:
            ctx.write_u8(f, col[0])
            ctx.write_string(f, col[1])

        ctx.write_s16(f, self.mac_font)
        ctx.write_s16(f, self.mac_font_size)
        ctx.write_s16(f, self.mac_image_scale)

        ctx.write_rect(f, self.home_rect)

        ctx.write_rgb_color(f, self.background_color)
        ctx.write_rgb_color(f, self.forground_color)

        ctx.write_s16(f, self.ql_image_scale)
        ctx.write_object_ref(self.root, f, self.attributes)
        ctx.write_bool(f, self.was_iconic)

        if self.class_id[:] == b'ABIN':
            ctx.write_u8(f, 0x03)

    def build_mob_dict(self):
        self.mob_dict = {}
        for mob in self.mobs:
            self.mob_dict[mob.mob_id] = mob

    def add_mob(self, mob):
        bin_item = self.root.create.BinItem()
        bin_item.mob = mob
        self.items.append(bin_item)

    @property
    def mobs(self):
        for item in self.items:
            yield item.mob

    def toplevel(self):
        for mob in self.mobs:
            if  mob.mob_type in ('CompositionMob', ) and mob.usage_code == 0:
                yield mob

    def compositionmobs(self):
        for mob in self.mobs:
            if  mob.mob_type in ('CompositionMob', ):
                yield mob

    def mastermobs(self):
        for mob in self.mobs:
            if  mob.mob_type in ('MasterMob', ):
                yield mob


@utils.register_class
class BinFirst(Bin):
    class_id = b'BINF'
    propertydefs_dict = {}
    propertydefs = Bin.propertydefs + [
        AVBPropertyDef('unknown_s32', 'unknown_s32', 'int32',  0),
    ]
    __slots__ = ()

    def read(self, f):
        super(BinFirst, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        self.unknown_s32 = ctx.read_s32(f)

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(BinFirst, self).write(f)
        ctx = self.root.octx

        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)

        ctx.write_s32(f, self.unknown_s32)

        ctx.write_u8(f, 0x03)



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )

from . import core
from . import utils
from .core import AVBPropertyDef, AVBRefList
from . import mobid
from . utils import peek_data

class FileLocator(core.AVBObject):
    class_id = b'FILE'
    propertydefs_dict = {}
    propertydefs = [
        AVBPropertyDef('path',        'OMFI:FL:PathName',       'string'),
        AVBPropertyDef('path_posix',  'OMFI:FL:POSIXPathName',  'string'),
        AVBPropertyDef('path_utf8',   'OMFI:FL:PathNameUTF8',   'string'),
        AVBPropertyDef('path2_utf8',  'OMFI:FL:PathNameUTF8',   'string')
    ]
    __slots__ = ()

    def read(self, f):
        super(FileLocator, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x02)

        self.path = ctx.read_string(f)

        for tag in ctx.iter_ext(f):
            if tag == 0x01:
                ctx.read_assert_tag(f, 76)
                self.path_posix = ctx.read_string(f)
            elif tag == 0x02:
                ctx.read_assert_tag(f, 76)
                self.path_utf8 = ctx.read_string(f, 'utf-8')
            elif tag == 0x03:
                ctx.read_assert_tag(f, 76)
                self.path2_utf8 = ctx.read_string(f, 'utf-8')
            else:
                raise ValueError("%s: unknown ext tag 0x%02X %d" % (str(self.class_id), tag,tag))

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(FileLocator, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x02)

        ctx.write_string(f, self.path)

        if hasattr(self, 'path_posix'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 76)
            ctx.write_string(f, self.path_posix)

        if hasattr(self, 'path_utf8'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x02)
            ctx.write_u8(f, 76)
            ctx.write_string(f, self.path_utf8, encoding='utf-8')

        if hasattr(self, 'path2_utf8'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x03)
            ctx.write_u8(f, 76)
            ctx.write_string(f, self.path2_utf8, encoding='utf-8')

        ctx.write_u8(f, 0x03)

@utils.register_class
class MacFileLocator(FileLocator):
    class_id = b'FILE'
    propertydefs_dict = {}
    __slots__ = ()

@utils.register_class
class WinFileLocator(FileLocator):
    class_id = b'WINF'
    propertydefs_dict = {}
    __slots__ = ()

@utils.register_class
class URLLocator(core.AVBObject):
    class_id = b'URLL'
    propertydefs_dict = {}
    __slots__ = ()

    def read(self, f):
        super(URLLocator, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(URLLocator, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x03)

@utils.register_class
class GraphicEffect(core.AVBObject):
    class_id = b'GRFX'
    propertydefs_dict = {}
    propertydefs = [
        AVBPropertyDef('pict_data', 'OMFI:MC:GRFX:PictData', 'bytes'),
    ]
    __slots__ = ()

    def read(self, f):
        super(GraphicEffect, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        pict_size = ctx.read_s32(f)
        assert pict_size >= 0

        self.pict_data = bytearray(f.read(pict_size))
        assert len(self.pict_data) == pict_size

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(GraphicEffect, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)

        ctx.write_s32(f, len(self.pict_data))
        f.write(self.pict_data)

        ctx.write_u8(f, 0x03)

@utils.register_class
class ShapeList(core.AVBObject):
    class_id = b'SHLP'
    propertydefs_dict = {}
    propertydefs = [
        AVBPropertyDef('shape_data', 'ShapeList', 'bytes'),
    ]

    def read(self, f):
        super(ShapeList, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        data_size = ctx.read_s32(f)
        assert data_size >= 0

        self.shape_data = bytearray(f.read(data_size))
        assert len(self.shape_data) == data_size
        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(ShapeList, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)

        ctx.write_s32(f, len(self.shape_data))
        f.write(self.shape_data)

        ctx.write_u8(f, 0x03)

@utils.register_class
class ColorCorrectionEffect(core.AVBObject):
    class_id = b'CCFX'
    propertydefs_dict = {}
    propertydefs = [
        AVBPropertyDef('color_correction', 'OMFI:FXPS:colorCorrection', 'bytes'),
    ]
    __slots__ = ()

    def read(self, f):
        super(ColorCorrectionEffect, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        data_size = ctx.read_s16(f)
        assert data_size >= 0
        #
        self.color_correction = bytearray(f.read(data_size))
        assert len(self.color_correction) == data_size
        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(ColorCorrectionEffect, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)

        ctx.write_s16(f, len(self.color_correction))
        f.write(self.color_correction)

        ctx.write_u8(f, 0x03)

@utils.register_helper_class
class EffectParam(core.AVBObject):
    propertydefs_dict = {}
    propertydefs = [
        AVBPropertyDef('percent_time',     'OMFI:FXPS:percentTime',         'int32'),
        AVBPropertyDef('level',            'OMFI:FXPS:level',               'int32'),
        AVBPropertyDef('pos_x',            'OMFI:FXPS:posX',                'int32'),
        AVBPropertyDef('floor_x',          'OMFI:FXPS:xFloor',              'int32'),
        AVBPropertyDef('ceil_x',           'OMFI:FXPS:xCeiling',            'int32'),
        AVBPropertyDef('pos_y',            'OMFI:FXPS:posY',                'int32'),
        AVBPropertyDef('floor_y',          'OMFI:FXPS:yFloor',              'int32'),
        AVBPropertyDef('ceil_y',           'OMFI:FXPS:yCeiling',            'int32'),
        AVBPropertyDef('scale_x',          'OMFI:FXPS:xScale',              'int32'),
        AVBPropertyDef('scale_y',          'OMFI:FXPS:yScale',              'int32'),
        AVBPropertyDef('crop_left',        'OMFI:FXPS:cropLeft',            'int32'),
        AVBPropertyDef('crop_right',       'OMFI:FXPS:cropRight',           'int32'),
        AVBPropertyDef('crop_top',         'OMFI:FXPS:cropTop',             'int32'),
        AVBPropertyDef('crop_bottom',      'OMFI:FXPS:cropBottom',          'int32'),
        AVBPropertyDef('box',              'OMFI:FXPS:box',                 'list'),
        AVBPropertyDef('box_xscale',       'OMFI:FXPS:boxLvl2Xscale',       'bool'),
        AVBPropertyDef('box_yscale',       'OMFI:FXPS:boxLvl2Yscale',       'bool'),
        AVBPropertyDef('box_xpos',         'OMFI:FXPS:FXboxLvl2Xpos',       'bool'),
        AVBPropertyDef('box_ypos',         'OMFI:FXPS:omFXboxLvl2Ypos',     'bool'),
        AVBPropertyDef('border_width',     'OMFI:FXPS:borderWidth',         'int32'),
        AVBPropertyDef('border_soft',      'OMFI:FXPS:borderSoft',          'int32'),
        AVBPropertyDef('splill_gain2',     'OMFI:FXPS:spillSecondGain',     'int16'),
        AVBPropertyDef('splill_gain',      'OMFI:FXPS:spillGain',           'int16'),
        AVBPropertyDef('splill_soft2',     'OMFI:FXPS:spillSecondSoft',     'int16'),
        AVBPropertyDef('splill_soft',      'OMFI:FXPS:spillSoft',           'int16'),
        AVBPropertyDef('enable_key_flags', 'OMFI:FXPS:enableKeyFlags',      'int8'),
        AVBPropertyDef('colors',           'OMFI:FXPS:Colors',              'list'),
        AVBPropertyDef('user_param',       'OMFI:FXPS:userParam',           'bytes'),
        AVBPropertyDef('selected',         'OMFI:FXPS:selected',            'bool'),
    ]
    __slots__ = ()

@utils.register_class
class EffectParamList(core.AVBObject):
    class_id = b'FXPS'
    propertydefs_dict = {}
    propertydefs = [
        AVBPropertyDef('orig_length',   'OMFI:FXPS:originalLength',   'int32'),
        AVBPropertyDef('window_offset', 'OMFI:FXPS:omFXwindowOffset', 'int32'),
        AVBPropertyDef('keyframe_size', 'OMFI:FXPS:keyFrameSize',     'int32'),
        AVBPropertyDef('parameters',   'paramamters',                 'list'),
    ]
    __slots__ = ()

    def read(self, f):
        super(EffectParamList, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x12)

        self.orig_length = ctx.read_s32(f)
        self.window_offset = ctx.read_s32(f)

        count = ctx.read_s32(f)
        self.keyframe_size = ctx.read_s32(f)
        self.parameters = []

        for i in range(count):
            p = EffectParam.__new__(EffectParam, root=self.root)
            start = f.tell()
            p.percent_time = ctx.read_s32(f)
            p.level = ctx.read_s32(f)
            p.pos_x = ctx.read_s32(f)
            p.floor_x = ctx.read_s32(f)
            p.ceil_x = ctx.read_s32(f)
            p.pos_y = ctx.read_s32(f)
            p.floor_y = ctx.read_s32(f)
            p.ceil_y = ctx.read_s32(f)
            p.scale_x = ctx.read_s32(f)
            p.scale_y = ctx.read_s32(f)

            p.crop_left = ctx.read_s32(f)
            p.crop_right = ctx.read_s32(f)
            p.crop_top = ctx.read_s32(f)
            p.crop_bottom = ctx.read_s32(f)

            p.box = []
            # boxTop
            p.box.append(ctx.read_s32(f))
            # boxBottom
            p.box.append(ctx.read_s32(f))
            # boxTop repeat??
            p.box.append(ctx.read_s32(f))
            # boxRight
            p.box.append(ctx.read_s32(f))

            p.box_xscale = ctx.read_bool(f)
            p.box_yscale = ctx.read_bool(f)
            p.box_xpos = ctx.read_bool(f)
            p.box_ypos = ctx.read_bool(f)

            p.border_width = ctx.read_s32(f)
            p.border_soft = ctx.read_s32(f)

            p.splill_gain2 = ctx.read_s16(f)
            p.splill_gain = ctx.read_s16(f)
            p.splill_soft2 = ctx.read_s16(f)
            p.splill_soft = ctx.read_s16(f)

            p.enable_key_flags = ctx.read_s8(f)

            p.colors = []
            color_count = ctx.read_s32(f)
            assert color_count >= 0
            for j in range(color_count):
                color = ctx.read_s32(f)
                p.colors.append(color)

            param_size = ctx.read_s32(f)
            assert param_size >= 0

            p.user_param = bytearray(f.read(param_size))
            p.selected = ctx.read_bool(f)

            self.parameters.append(p)

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(EffectParamList, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x12)

        ctx.write_s32(f, self.orig_length)
        ctx.write_s32(f, self.window_offset)

        ctx.write_s32(f, len(self.parameters))
        ctx.write_s32(f, self.keyframe_size)

        for p in self.parameters:

            ctx.write_s32(f, p.percent_time)
            ctx.write_s32(f, p.level)
            ctx.write_s32(f, p.pos_x)
            ctx.write_s32(f, p.floor_x)
            ctx.write_s32(f, p.ceil_x)
            ctx.write_s32(f, p.pos_y)
            ctx.write_s32(f, p.floor_y)
            ctx.write_s32(f, p.ceil_y)
            ctx.write_s32(f, p.scale_x)
            ctx.write_s32(f, p.scale_y)

            ctx.write_s32(f, p.crop_left)
            ctx.write_s32(f, p.crop_right)
            ctx.write_s32(f, p.crop_top)
            ctx.write_s32(f, p.crop_bottom)

            # boxTop
            ctx.write_s32(f, p.box[0])
            # boxBottom
            ctx.write_s32(f, p.box[1])
            # boxTop repeat??
            ctx.write_s32(f, p.box[2])
            # boxRight
            ctx.write_s32(f, p.box[3])

            ctx.write_bool(f, p.box_xscale)
            ctx.write_bool(f, p.box_yscale)
            ctx.write_bool(f, p.box_xpos)
            ctx.write_bool(f, p.box_ypos)

            ctx.write_s32(f, p.border_width)
            ctx.write_s32(f, p.border_soft)

            ctx.write_s16(f, p.splill_gain2)
            ctx.write_s16(f, p.splill_gain)
            ctx.write_s16(f, p.splill_soft2)
            ctx.write_s16(f, p.splill_soft)

            ctx.write_s8(f, p.enable_key_flags)

            ctx.write_s32(f, len(p.colors))

            for color in p.colors:
                ctx.write_s32(f, color)

            ctx.write_s32(f, len(p.user_param))
            f.write(p.user_param)
            ctx.write_bool(f, p.selected)

        ctx.write_u8(f, 0x03)

@utils.register_class
class CFUserParam(core.AVBObject):
    class_id = b'AVUP'
    propertydefs_dict = {}
    propertydefs = [
        AVBPropertyDef('byte_order', 'OMFI:AVUP:ByteOrder', 'uint16'),
        AVBPropertyDef('uuid',       'OMFI:AVUP:TypeID',    'UUID'),
        AVBPropertyDef('data',       'OMFI:AVUP:ValueData', 'bytes'),
    ]
    __slots__ = ()

    def read(self, f):
        super(CFUserParam, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        self.byte_order = ctx.read_s16(f)
        assert self.byte_order == 0x4949

        self.uuid = ctx.read_raw_uuid(f)

        # why twice?
        value_size1 = ctx.read_s32(f)
        value_size2 = ctx.read_s32(f)

        assert value_size2 == (value_size1 - 4)

        self.data = bytearray(f.read(value_size2))

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(CFUserParam, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)

        ctx.write_s16(f, 0x4949)

        ctx.write_raw_uuid(f, self.uuid)

        ctx.write_s32(f, len(self.data) + 4)
        ctx.write_s32(f, len(self.data))

        f.write(self.data)

        ctx.write_u8(f, 0x03)

@utils.register_class
class ParameterItems(core.AVBObject):
    class_id = b'PRIT'
    propertydefs_dict = {}
    propertydefs = [
        AVBPropertyDef('uuid',            'OMFI:PRIT:GUID',                   'UUID'),
        AVBPropertyDef('value_type',      'OMFI:PRIT:ValueType',              'int16'),
        AVBPropertyDef('value',           'OMFI:PRIT:Value',                  'int32'),
        AVBPropertyDef('name',            'OMFI:PRIT:Name',                   'string'),
        AVBPropertyDef('enable',          'OMFI:PRIT:Enabled',                'bool'),
        AVBPropertyDef('control_track',   'OMFI:PRIT:ControlTrack',           'reference'),
        AVBPropertyDef('contribs_to_sig', 'OMFI:PRIT:ContributesToSignature', 'bool'),
    ]
    __slots__ = ()

    def read(self, f):
        super(ParameterItems, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x02)

        self.uuid = ctx.read_raw_uuid(f)
        self.value_type = ctx.read_s16(f)
        if self.value_type == 1:
            self.value = ctx.read_s32(f)
        elif self.value_type == 2:
            self.value = ctx.read_double(f)
        elif self.value_type == 4:
            self.value = ctx.read_object_ref(self.root, f)
        else:
            raise ValueError("unknown value_type: %d" % self.value_type)

        self.name = ctx.read_string(f)
        self.enable = ctx.read_bool(f)
        self.control_track = ctx.read_object_ref(self.root, f)

        for tag in ctx.iter_ext(f):
            if tag == 0x01:
                ctx.read_assert_tag(f, 66)
                self.contribs_to_sig = ctx.read_bool(f)
            else:
                raise ValueError("%s: unknown ext tag 0x%02X %d" % (str(self.class_id), tag,tag))

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(ParameterItems, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x02)

        ctx.write_raw_uuid(f, self.uuid)
        ctx.write_s16(f, self.value_type)

        if self.value_type == 1:
            ctx.write_s32(f, self.value)
        elif self.value_type == 2:
            ctx.write_double(f, self.value)
        elif self.value_type == 4:
            ctx.write_object_ref(self.root, f, self.value)
        else:
            raise ValueError("unknown value_type: %d" % self.value_type)

        if self.name:
            ctx.write_string(f, self.name)
        else:
            ctx.write_u16(f, 0xFFFF)

        ctx.write_bool(f, self.enable)
        ctx.write_object_ref(self.root, f, self.control_track)

        if hasattr(self, 'contribs_to_sig'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 66)
            ctx.write_bool(f, self.contribs_to_sig)

        ctx.write_u8(f, 0x03)

@utils.register_class
class MSMLocator(core.AVBObject):
    class_id = b'MSML'
    propertydefs_dict = {}
    propertydefs = [
        AVBPropertyDef('last_known_volume',        'OMFI:MSML:LastKnownVolume',      'string'),
        AVBPropertyDef('domain_type',              'OMFI:MSML:DomainType',           'int32'),
        AVBPropertyDef('mob_id',                   'MobID',                          'MobID'),
        AVBPropertyDef('last_known_volume_utf8',   'OMFI:MSML:LastKnownVolumeUTF8', 'string'),
    ]
    __slots__ = ()

    def read(self, f):
        super(MSMLocator, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x02)

        mob_id_hi = ctx.read_u32(f)
        mob_id_lo = ctx.read_u32(f)

        self.last_known_volume = ctx.read_string(f)

        for tag in ctx.iter_ext(f):

            if tag == 0x01:
                ctx.read_assert_tag(f, 71)
                self.domain_type = ctx.read_s32(f)
            elif tag == 0x02:
                self.mob_id = ctx.read_mob_id(f)
            elif tag == 0x03:
                ctx.read_assert_tag(f, 76)
                self.last_known_volume_utf8 = ctx.read_string(f, 'utf-8')
            else:
                raise ValueError("%s: unknown ext tag 0x%02X %d" % (str(self.class_id), tag,tag))

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(MSMLocator, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x02)

        lo = self.mob_id.material.time_low
        hi = self.mob_id.material.time_mid + (self.mob_id.material.time_hi_version << 16)
        ctx.write_u32(f, lo)
        ctx.write_u32(f, hi)

        ctx.write_string(f, self.last_known_volume)

        if hasattr(self, 'domain_type'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.domain_type)

        if hasattr(self, 'mob_id'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x02)
            ctx.write_mob_id(f, self.mob_id)

        if hasattr(self, 'last_known_volume_utf8'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x03)
            ctx.write_u8(f, 76)
            ctx.write_string(f, self.last_known_volume_utf8, 'utf-8')

        ctx.write_u8(f, 0x03)

@utils.register_class
class Position(core.AVBObject):
    class_id = b'APOS'
    propertydefs_dict = {}
    propertydefs = [
        AVBPropertyDef('mob_id', "MobID", 'MobID'),
    ]
    __slots__ = ()

    def read(self, f):
        super(Position, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        mob_id_hi = ctx.read_u32(f)
        mob_id_lo = ctx.read_u32(f)

        for tag in ctx.iter_ext(f):
            if tag == 0x01:
                self.mob_id = ctx.read_mob_id(f)
            else:
                raise ValueError("%s: unknown ext tag 0x%02X %d" % (str(self.class_id), tag,tag))

        if self.class_id[:] ==  b'APOS':
            ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(Position, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)

        lo = self.mob_id.material.time_low
        hi = self.mob_id.material.time_mid + (self.mob_id.material.time_hi_version << 16)
        ctx.write_u32(f, lo)
        ctx.write_u32(f, hi)

        if hasattr(self, 'mob_id'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x01)
            ctx.write_mob_id(f, self.mob_id)

        if self.class_id[:] ==  b'APOS':
            ctx.write_u8(f, 0x03)

@utils.register_class
class BOBPosition(Position):
    class_id = b'ABOB'
    propertydefs_dict = {}
    propertydefs = Position.propertydefs + [
        AVBPropertyDef('sample_num',  "__OMFI:MSBO:sampleNum",   'int32'),
        AVBPropertyDef('length',      "__OMFI:MSBO:length",      'int32'),
        AVBPropertyDef('track_type',  "OMFI:trkt:Track.trkType", 'int32'),
        AVBPropertyDef('track_index', "OMFI:trkt:Track.trkLNum", 'int32'),
    ]
    __slots__ = ()

    def read(self, f):
        super(BOBPosition, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        self.sample_num = ctx.read_s32(f)
        self.length = ctx.read_s32(f)
        self.track_type = ctx.read_s16(f)
        self.track_index = ctx.read_s16(f)

        if self.class_id[:] ==  b'ABOB':
            ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(BOBPosition, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)

        ctx.write_s32(f, self.sample_num)
        ctx.write_s32(f, self.length)
        ctx.write_s16(f, self.track_type)
        ctx.write_s16(f, self.track_index)

        if self.class_id[:] ==  b'ABOB':
            ctx.write_u8(f, 0x03)

@utils.register_class
class DIDPosition(BOBPosition):
    class_id = b'DIDP'
    propertydefs_dict = {}
    propertydefs = BOBPosition.propertydefs + [
        AVBPropertyDef('strip',        "_Strip",       'int32'),
        AVBPropertyDef('offset',       "_Offset",      'uint64'),
        AVBPropertyDef('byte_length',  "_ByteLength",  'uint64'),
        AVBPropertyDef('spos_invalid', "_SPosInvalid", 'bool'),
    ]
    __slots__ = ()

    def read(self, f):
        super(DIDPosition, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        self.strip = ctx.read_s32(f)
        self.offset = ctx.read_u64(f)
        self.byte_length = ctx.read_u64(f)
        self.spos_invalid = ctx.read_bool(f)

        if self.class_id[:] == b'DIDP':
            ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(DIDPosition, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)

        ctx.write_s32(f, self.strip)
        ctx.write_u64(f, self.offset)
        ctx.write_u64(f, self.byte_length)
        ctx.write_bool(f, self.spos_invalid)

        if self.class_id[:] == b'DIDP':
            ctx.write_u8(f, 0x03)

@utils.register_class
class MPGPosition(DIDPosition):
    class_id = b'MPGP'
    propertydefs_dict = {}
    propertydefs = DIDPosition.propertydefs + [
     AVBPropertyDef('trailing_discards',    '_trailingDiscards',      'int16'),
     AVBPropertyDef('need_seq_hdr',          '_needSeqHdr',            'Boolean'),
     AVBPropertyDef('leader_length',         'leaderLength',           'int16'),
     AVBPropertyDef('fields', '               leadingDiscardFields',   'list'),
    ]
    __slots__ = ()

    def read(self, f):
        super(MPGPosition, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        self.trailing_discards = ctx.read_s16(f)
        self.need_seq_hdr = ctx.read_bool(f)
        self.fields = []

        # NOTE: I think leading_discard_fields == 2 * leader_length
        leader_length = ctx.read_s16(f)
        if leader_length > 0:
            leading_discard_fields = ctx.read_s16(f)
            for i in range(leader_length):
                picture_type = ctx.read_u8(f)
                length = ctx.read_u32(f)
                self.fields.append([picture_type, length])

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(MPGPosition, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)

        ctx.write_s16(f, self.trailing_discards)
        ctx.write_bool(f, self.need_seq_hdr)

        leader_length = len(self.fields)
        ctx.write_s16(f, leader_length)
        if leader_length > 0:
            ctx.write_s16(f, leader_length*2)

            for picture_type, length in self.fields:
                ctx.write_u8(f, picture_type)
                ctx.write_u32(f, length)

        ctx.write_u8(f, 0x03)

@utils.register_class
class BinRef(core.AVBObject):
    class_id = b'MCBR'
    propertydefs_dict = {}
    propertydefs = [
        AVBPropertyDef('uid_high',  'OMFI:MCBR:MC:binID.high',  'int32'),
        AVBPropertyDef('uid_low',   'OMFI:MCBR:MC:binID.low',   'int32'),
        AVBPropertyDef('name',      'OMFI:MCBR:MC:binName',     'string'),
        AVBPropertyDef('name_utf8', 'OMFI:MCBR:MC:binNameUTF8', 'string'),
    ]
    __slots__ = ()

    def read(self, f):
        super(BinRef, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        self.uid_high = ctx.read_s32(f)
        self.uid_low = ctx.read_s32(f)
        self.name = ctx.read_string(f)

        for tag in ctx.iter_ext(f):
            if tag == 0x01:
                ctx.read_assert_tag(f, 76)
                self.name_utf8 = ctx.read_string(f,'utf-8')
            else:
                raise ValueError("%s: unknown ext tag 0x%02X %d" % (str(self.class_id), tag,tag))

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(BinRef, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)

        ctx.write_s32(f, self.uid_high)
        ctx.write_s32(f, self.uid_low)
        ctx.write_string(f, self.name)

        ctx.write_u8(f, 0x01)
        ctx.write_u8(f, 0x01)
        ctx.write_u8(f, 76)
        ctx.write_string(f, self.name_utf8, 'utf-8')

        ctx.write_u8(f, 0x03)

@utils.register_class
class MobRef(core.AVBObject):
    class_id = b'MCMR'
    propertydefs_dict = {}
    propertydefs = [
            AVBPropertyDef('position',      'OMFI:MCMR:MC:Position', 'int32'),
            AVBPropertyDef('mob_id',        'MobID', 'MobID'),
    ]
    __slots__ = ()

    def read(self, f):
        super(MobRef, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        mob_hi = ctx.read_u32(f)
        mob_lo = ctx.read_u32(f)
        self.position = ctx.read_s32(f)

        for tag in ctx.iter_ext(f):
            if tag == 0x01:
                self.mob_id = ctx.read_mob_id(f)
            else:
                raise ValueError("%s: unknown ext tag 0x%02X %d" % (str(self.class_id), tag,tag))

        if self.class_id[:] == b'MCMR':
            ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(MobRef, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)

        lo = self.mob_id.material.time_low
        hi = self.mob_id.material.time_mid + (self.mob_id.material.time_hi_version << 16)

        ctx.write_u32(f, lo)
        ctx.write_u32(f, hi)
        ctx.write_s32(f, self.position)

        if hasattr(self, 'mob_id'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x01)
            ctx.write_mob_id(f, self.mob_id)

        if self.class_id[:] == b'MCMR':
            ctx.write_u8(f, 0x03)

# also called a TimeCrumb
@utils.register_class
class Marker(MobRef):
    class_id = b'TMBC'
    propertydefs_dict = {}
    propertydefs = MobRef.propertydefs + [
        AVBPropertyDef('comp_offset',   'OMFI:TMBC:MC:CompOffset',             'int32'),
        AVBPropertyDef('attributes',    'OMFI:TMBC:MC:Attributes',             'reference'),
        AVBPropertyDef('color',         'OMFI:TMBC:MC:CarbonAPI::RGBColor',    'list'),
        AVBPropertyDef('handled_codes', 'OMFI:TMBC:MC:handledBadControlCodes', 'bool'),
    ]
    __slots__ = ()

    def read(self, f):
        super(Marker, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x03)

        self.comp_offset = ctx.read_s32(f)
        self.attributes = ctx.read_object_ref(self.root, f)
        # print(self.comp_offset, self.attributes)

        version = ctx.read_s16(f)
        assert version == 1

        self.color = []
        self.color.append(ctx.read_u16(f))
        self.color.append(ctx.read_u16(f))
        self.color.append(ctx.read_u16(f))

        for tag in ctx.iter_ext(f):

            if tag == 0x01:
                ctx.read_assert_tag(f, 66)
                self.handled_codes = ctx.read_bool(f)
            else:
                raise ValueError("%s: unknown ext tag 0x%02X %d" % (str(self.class_id), tag,tag))

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(Marker, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x03)

        ctx.write_s32(f, self.comp_offset)
        ctx.write_object_ref(self.root, f, self.attributes)

        #version
        ctx.write_s16(f, 1)

        for c in self.color:
            ctx.write_u16(f, c)

        if hasattr(self, 'handled_codes'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 66)
            ctx.write_bool(f ,self.handled_codes)

        ctx.write_u8(f, 0x03)

@utils.register_class
class TrackerManager(core.AVBObject):
    class_id = b'TKMN'
    propertydefs_dict = {}
    propertydefs = [
        AVBPropertyDef('data_slots',  'OMFI:TKMN:TrackerDataSlots',  'reference'),
        AVBPropertyDef('param_slots', 'OMFI:TKMN:TrackedParamSlots', 'reference'),
    ]
    __slots__ = ()

    def read(self, f):
        super(TrackerManager, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        self.data_slots = ctx.read_object_ref(self.root, f)
        self.param_slots = ctx.read_object_ref(self.root, f)

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(TrackerManager, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)

        ctx.write_object_ref(self.root, f, self.data_slots)
        ctx.write_object_ref(self.root, f, self.param_slots)

        ctx.write_u8(f, 0x03)

@utils.register_class
class TrackerDataSlot(core.AVBObject):
    class_id = b'TKDS'
    propertydefs_dict = {}
    propertydefs = [
        AVBPropertyDef('tracker_data',  'OMFI:TKDS:TrackerData',      'ref_list'),
        AVBPropertyDef('track_fg',      'OMFI:TKDAS:TrackForeground', 'bool'),
    ]
    __slots__ = ()

    def read(self, f):
        super(TrackerDataSlot, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        count = ctx.read_s32(f)
        self.tracker_data = AVBRefList.__new__(AVBRefList, root=self.root)
        for i in range(count):
            ref = ctx.read_object_ref(self.root, f)
            self.tracker_data.append(ref)

        for tag in ctx.iter_ext(f):
            if tag == 0x01:
                ctx.read_assert_tag(f, 66)
                self.track_fg = ctx.read_bool(f)
            else:
                raise ValueError("%s: unknown ext tag 0x%02X %d" % (str(self.class_id), tag,tag))

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(TrackerDataSlot, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)

        ctx.write_s32(f, len(self.tracker_data))

        for track in self.tracker_data:
            ctx.write_object_ref(self.root, f, track)

        if hasattr(self, 'track_fg'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 66)
            ctx.write_bool(f, self.track_fg)

        ctx.write_u8(f, 0x03)

@utils.register_class
class TrackerParameterSlot(core.AVBObject):
    class_id = b'TKPS'
    propertydefs_dict = {}
    propertydefs = [
        AVBPropertyDef('settings', 'OMFI:TKPS:EffectSettings', 'bytes'),
        AVBPropertyDef('params',   'OMFI:TKPS:TrackedParam',   'ref_list'),
    ]
    __slots__ = ()

    def read(self, f):
        super(TrackerParameterSlot, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)
        size = ctx.read_s16(f)
        assert size >= 0
        self.settings = bytearray(f.read(size))

        count = ctx.read_s32(f)
        assert count >= 0
        self.params = AVBRefList.__new__(AVBRefList, root=self.root)
        for i in range(count):
            ref = ctx.read_object_ref(self.root, f)
            self.params.append(ref)

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(TrackerParameterSlot, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)

        ctx.write_s16(f, len(self.settings))
        f.write(self.settings)

        ctx.write_s32(f, len(self.params))
        for p in self.params:
            ctx.write_object_ref(self.root, f, p)

        ctx.write_u8(f, 0x03)

@utils.register_class
class TrackerData(core.AVBObject):
    class_id = b'TKDA'
    propertydefs_dict = {}
    propertydefs = [
        AVBPropertyDef('settings',        'OMFI:TKDA:TrackerSettings',            'bytes'),
        AVBPropertyDef('clip_version',    'OMFI:TKDA:TrackerClipVersion',         'uint32'),
        AVBPropertyDef('clips',           'OMFI:TKDA:TrackerClip',                'ref_list'),
        AVBPropertyDef('offset_tracking', 'OMFI:TKDA:TrackerOffsetTracking',      'uint32'),
        AVBPropertyDef('smoothing',       'OMFI:TKDA:TrackerSmoothing',           'uint32'),
        AVBPropertyDef('jitter_removal',  'name="OMFI:TKDA:TrackerJitterRemoval', 'uint32'),
        AVBPropertyDef('filter_amount',  'name="OMFI:TKDA:TrackerFilterDataAmt',  'double'),
        AVBPropertyDef('clip5',           'OMFI:TKDA:TrackerClip',                'reference'),
        AVBPropertyDef('clip6',           'OMFI:TKDA:TrackerClip',                'reference'),
    ]
    __slots__ = ()

    def read(self, f):
        super(TrackerData, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        setting_size = ctx.read_s16(f)
        self.settings = bytearray(f.read(setting_size))
        self.clip_version = ctx.read_u32(f)

        count = ctx.read_s16(f)
        self.clips = AVBRefList.__new__(AVBRefList, root=self.root)
        for i in range(count):
            ref = ctx.read_object_ref(self.root, f)
            self.clips.append(ref)

        for tag in ctx.iter_ext(f):
            if tag == 0x01:
                ctx.read_assert_tag(f, 72)
                self.offset_tracking = ctx.read_u32(f)
            elif tag == 0x02:
                ctx.read_assert_tag(f, 72)
                self.smoothing = ctx.read_u32(f)
            elif tag == 0x03:
                ctx.read_assert_tag(f, 72)
                self.jitter_removal = ctx.read_u32(f)
            elif tag == 0x04:
                ctx.read_assert_tag(f, 75)
                self.filter_amount = ctx.read_double(f)
            elif tag == 0x05:
                ctx.read_assert_tag(f, 72)
                self.clip5 = ctx.read_object_ref(self.root, f)
            elif tag == 0x06:
                ctx.read_assert_tag(f, 72)
                self.clip6 = ctx.read_object_ref(self.root, f)
            else:
                raise ValueError("%s: unknown ext tag 0x%02X %d" % (str(self.class_id), tag,tag))

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(TrackerData, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)

        ctx.write_s16(f, len(self.settings))
        f.write(self.settings)

        ctx.write_u32(f, self.clip_version)

        ctx.write_s16(f, len(self.clips))

        for clip in self.clips:
            ctx.write_object_ref(self.root, f, clip)

        if hasattr(self, 'offset_tracking'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 72)
            ctx.write_u32(f, self.offset_tracking)

        if hasattr(self, 'smoothing'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x02)
            ctx.write_u8(f, 72)
            ctx.write_u32(f, self.smoothing)

        if hasattr(self, 'jitter_removal'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x03)
            ctx.write_u8(f, 72)
            ctx.write_u32(f, self.jitter_removal)

        if hasattr(self, 'filter_amount'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x04)
            ctx.write_u8(f, 75)
            ctx.write_double(f, self.filter_amount)

        if hasattr(self, 'clip5'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x05)
            ctx.write_u8(f, 72)
            ctx.write_object_ref(self.root, f, self.clip5)

        if hasattr(self, 'clip6'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x06)
            ctx.write_u8(f, 72)
            ctx.write_object_ref(self.root, f, self.clip6)

        ctx.write_u8(f, 0x03)


@utils.register_class
class TrackerParameter(core.AVBObject):
    class_id = b'TKPA'
    propertydefs_dict = {}
    propertydefs = [
        AVBPropertyDef('settings', 'OMFI:TKPA:ParamSettings','bytes'),
    ]
    __slots__ = ()

    def read(self, f):
        super(TrackerParameter, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)
        size = ctx.read_s16(f)
        assert size >= 0
        self.settings = bytearray(f.read(size))

        ctx.read_assert_tag(f, 0x03)


    def write(self, f):
        super(TrackerParameter, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)

        ctx.write_s16(f, len(self.settings))
        f.write(self.settings)

        ctx.write_u8(f, 0x03)
